/*
 *
 * SdbCompile.c --
 *
 *      This file contains compiler, expression evaluator, and
 *      other functions that support the run-time expression
 *      evaluator.  Expressions are processed first by the
 *      lexer and parser which are defined in other files.
 *
 * Copyright (c) 1995 Tom Moore.  All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all copies of this software.
 * 
 */

#include "sdbInt.h"
#include "math.h"

/* 
 * Definitions
 */
#define CYCLING 0
#define OK 1
#define TRUE 1
#define FALSE 0

static void                    free_expression _ANSI_ARGS_((SdbFile *dbPtr,
				       SdbVariable *vars));
static SdbNode                *tree_copy _ANSI_ARGS_((SdbNode *orig));
static SdbDependentList*       add_dependent _ANSI_ARGS_(( char *me, 
				       char *code, SdbDependentList *top, 
				       SdbFile *SdbPtr));
static SdbDependentList *      down_add _ANSI_ARGS_((SdbFile *dbPtr, 
				       SdbNode *e, SdbDependentList *top, 
				       char *start));
static SdbDependentList *      free_down_up _ANSI_ARGS_((SdbFile *dbPtr, 
				       char *id, SdbDependentList *down));
static int                     up_check _ANSI_ARGS_((SdbFile *dbPtr, 
				       SdbVariable *varPtr, 
				       SdbVariable *start));
static double                  evaluate _ANSI_ARGS_((SdbFile *dbPtr, int n, 
				       SdbNode *tree));
static char *                  charEval _ANSI_ARGS_((SdbFile *dbPtr, int n, 
				       SdbNode *tree));
int                            cdfnor_ _ANSI_ARGS_((int *which, float *p, 
				       float *x, float *mean, float *sd, 
				       int *status, float *bound));
static void                    reset_error_status _ANSI_ARGS_((SdbFile *dbPtr));
static int                     does_assignment _ANSI_ARGS_((SdbNode *node));


/*
 *------------------------------------------------------------
 *
 * SdbGetNode --
 *
 *      Allocates space for a new node.
 *
 * Results:
 *      Returns the address of the new node.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
SdbNode               * 
SdbGetNode _ANSI_ARGS_((void)) 
{
  SdbNode *newnode=(SdbNode*)ckalloc(sizeof(SdbNode));
  if (!newnode) {
    fprintf(stderr,"Out of memory\n");
    exit(1);
  }
  newnode->right=NULL;
  newnode->left=NULL;
  return newnode;
}

/*
 *------------------------------------------------------------
 *
 * SdbFreeNode --
 *
 *      Frees a node and everything it contains
 *
 * Results:
 *      Returns NULL.
 *
 * Side effects:
 *      All nodes below this node are eliminated..
 *
 *------------------------------------------------------------
 */
SdbNode                *
SdbFreeNode(tree)
     SdbNode                *tree;
{
    if (!tree)
        return NULL;
    tree->right=SdbFreeNode(tree->right);
    tree->left=SdbFreeNode(tree->left);
    switch (tree->type) {
    case hsgCodeList:
        ckfree(tree->content.list.pntr);
        break;

    case chrChrFnct:
    case chrNumFnct:
    case numChrFnct:
    case numNumFnct:{
            SdbCat                 *pntr = tree->content.classList,
                                   *next;

            while (pntr) {
                pntr->expr=SdbFreeNode(pntr->expr);
                pntr->code=SdbFreeNode(pntr->code);
                next = pntr->next;
                ckfree(pntr);
                pntr = next;
            }
            break;
        }

    case curveFnct:{
            if (tree->content.yieldHandle->list)
	        ckfree(tree->content.yieldHandle->list);
            ckfree(tree->content.yieldHandle);
            break;
        }

    default:
        break;
    }
    ckfree(tree);

    return (NULL);
}


/*
 *------------------------------------------------------------
 *
 * free_expression --
 *
 *      Delete an expression that is stored in a variable slot.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Converts the variable to a constant of the same type.
 *
 *------------------------------------------------------------
 */
static void
free_expression(dbPtr, vars)
     SdbFile                *dbPtr;
     SdbVariable            *vars;
{

    if (!(vars->flag & SDB_EXPRESSION))
        return;

    vars->u.expr[0] = SdbFreeNode(vars->u.expr[0]);
    vars->u.expr[1] = SdbFreeNode(vars->u.expr[1]);
    vars->flag &= ~SDB_EXPRESSION;
    vars->flag |= SDB_CONSTANT;
    if (vars->flag && SDB_CODE) {
      int new;
      vars->u.code=GetIndices(dbPtr->htbl,".",&new);
    }
    else {
      vars->u.value=0.0;
    }
    dbPtr->stamp++;
    vars->down = free_down_up(dbPtr, vars->id, vars->down);
    vars->down = NULL;
}


/*
 *------------------------------------------------------------
 *
 * tree_copy --
 *
 *      make a duplicate of an expression tree
 *
 * Results:
 *      Pointer to the duplicate tree.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static SdbNode                *
tree_copy(orig)
     SdbNode                *orig;
{
    SdbNode                *new;

    if (!orig)
        return NULL;
    new = SdbGetNode();

    bcopy((char *) orig, (char *) new, sizeof(SdbNode));

    switch (orig->type) {
    case hsgCodeList:{
            int                     i;

            new->content.list.pntr =
                    (char **) ckalloc(new->content.list.n * sizeof(char *));

            for (i = 0; i < new->content.list.n; i++)
                new->content.list.pntr[i] = orig->content.list.pntr[i];
            break;
        }
    case chrChrFnct:
    case numChrFnct:
    case chrNumFnct:
    case numNumFnct:{
            SdbCat                **next,
                                   *e = orig->content.classList;

            next = &new->content.classList;

            while (e) {
                *next = (SdbCat *) ckalloc(sizeof(SdbCat));

                (*next)->expr = tree_copy(e->expr);
                (*next)->relation = e->relation;
                (*next)->code = tree_copy(e->code);
                (*next)->next = NULL;
                e = e->next;
                next = &((*next)->next);
            }
            break;
        }

    case curveFnct:{
            new->content.yieldHandle =
                    (SdbYieldHandle *) ckalloc(sizeof(SdbYieldHandle));

            bcopy((char *) orig->content.yieldHandle, 
		    (char *) new->content.yieldHandle,
                    sizeof(SdbYieldHandle));
	    new->content.yieldHandle->list=NULL;
	      

            break;
        }

    default:
        break;
    }

    new->right = tree_copy(orig->right);
    new->left = tree_copy(orig->left);
    return (new);
}

static SdbDependentList*
add_dependent(me, code, top, dbPtr)
     char *me;
     char *code;
     SdbDependentList *top;
     SdbFile *dbPtr;
{
  SdbDependentList       *temp,
                         *next = top;
  SdbVariable            *v;

  /*
   * This must be a new reference, so add it to the list.
   */
  temp = (SdbDependentList *) ckalloc(sizeof(SdbDependentList));
  temp->id = code;
  temp->next = top;
  top = temp;

  /* 
   * now add an up link from the other variable 
   */
  v = SdbFound(dbPtr->varPtr, code);
  if (v) {
    next = v->up;
    /*
     * Check the dependent variable for an existing up pointer. 
     */
    while (next) {
      if (next->id == me)
	return (top);
      next = next->next;
    }
    temp = (SdbDependentList *) ckalloc(sizeof(SdbDependentList));
    temp->id = me;
    temp->next = v->up;
    v->up = temp;
  }
  return(top);
}


/*
 *------------------------------------------------------------
 *
 * down_add --
 *
 *      Build a list of all variables that this expression
 *      refers to.  Add this variables names to the lists
 *      contained in the refered variables.
 *
 * Results:
 *      Pointer to the list.
 *
 * Side effects:
 *      Adds this variables name to other variables lists..
 *
 *------------------------------------------------------------
 */
static SdbDependentList *
down_add(dbPtr, e, top, start)
     SdbFile                *dbPtr;
     SdbNode                *e;
     SdbDependentList       *top;
     char                   *start;
{
     SdbDependentList       *next=top;

    /*
     * First check that this is a good node!
     */
    if (!e)
        return (top);

    /*
     * depending on the type of node, search for other variable pointers.
     */
    switch (e->type) {

    case noScan: {
      /* 
       * Don't scan below this part of the tree
       */
      return(top);
    }

      /*
       * References to lookup table values.  These will always 
       * depend on index variables.
       */
    case lookupNumber:
    case lookupCode: {
      SdbLookupHandle        *luHandlePtr=e->content.luHandlePtr;
      SdbLookup              *luPtr = luHandlePtr->luPtr;
      int i;

      for (i=0;i<luPtr->numIndex;i++) {
	next=top;
	while (next) {
	  if (!strcmp(luPtr->indexNames[i],next->id)) 
	    goto got_it;
	  next=next->next;
	}
	top=add_dependent(start ,luPtr->indexNames[i],top,dbPtr);
      got_it: ;
      }
      return(top);
    }

      /*
       * These are references to other variable names.  The node
       * contains the string name of the variable.
       */
    case hsgCode:
    case hsgNumber:{

	    /*
	     * Search the dependent list to see if this name
	     * is already present.  If so, return.
	     */
            while (next) {
                if (e->content.code == next->id)
                    return (top);
                next = next->next;
            }
	    top=add_dependent(start, e->content.code,top,dbPtr);
            return (top);
            break;
        }

      /*
       * If this is a code or numeric function, also descend the
       * expression structure.
       */
    case chrChrFnct:
    case numChrFnct:
    case chrNumFnct:
    case numNumFnct:{
            SdbCat                 *nextCat = e->content.classList;

            while (nextCat) {
                top = down_add(dbPtr, nextCat->expr, top, start);
                top = down_add(dbPtr, nextCat->code, top, start);
                nextCat = nextCat->next;
            }
            break;
        }

      /*
       * If an assignment, mark the assigned variable as mutable.
       */
    case assignChar:
    case assignNum: {
      SdbVariable *v;
      if ((v=SdbFound(dbPtr->varPtr,e->left->content.code))) 
	v->flag |= SDB_MUTABLE;
      break;
    }
      
    default:
        break;
    }

    /*
     * Finally, descend the left and right branches.
     */
    top = down_add(dbPtr, e->right, top, start);
    top = down_add(dbPtr, e->left, top, start);
    return (top);
}

/*
 * does_assignment --
 *
 * Search the expression tree looking for assignment nodes.
 *
 * Results:
 *   Return 1 if an assigment is made, 0 if none are found.
 *
 * Side effects:
 *   None
 */
static int
does_assignment(node)
     SdbNode *node;
{
  if (!node) 
    return(0);
  if ((node->type == assignNum) ||
      (node->type == assignChar))
    return(1);
  if (does_assignment(node->left))
    return(1);
  return(does_assignment(node->right));
}

/*
 *------------------------------------------------------------
 *
 * free_down_up --
 *
 *      Delete the dependent list.
 *
 * Results:
 *      NULL.
 *
 * Side effects:
 *      Deletes this variables name from other variables lists.
 *
 *------------------------------------------------------------
 */
static SdbDependentList *
free_down_up(dbPtr, id, down)
     SdbFile                *dbPtr;
     char                   *id;
     SdbDependentList       *down;
{
    SdbDependentList       *next = down;
    SdbVariable            *varPtr;

    while (next) {
        SdbDependentList       *temp;

        temp = next->next;
        if ((varPtr = SdbFound(dbPtr->varPtr, next->id))) {
            SdbDependentList      **up,
                                   *up_temp;

            up = &varPtr->up;
            while (*up) {
                if ((*up)->id == id) {
                    up_temp = *up;
                    *up = (*up)->next;
                    ckfree(up_temp);
                    break;
                }
                if (*up)
                    up = &((*up)->next);
            }
        }
        ckfree(next);
        next = temp;
    }
    return NULL;
}


/*
 *------------------------------------------------------------
 *
 * up_check --
 *
 * Check the dependency graph that describes the links between expressions.
 * The down list simply describes the variables that this expression 
 * depends on.  Each expression also has an up list, descibing which
 * other expressions are dependent on itself.  Ascend the list searching
 * for variable references back to the inital starting point.
 *
 * Results:
 *      A status code is returned which is OK usually, 
 *      or CYCLING if the expressions are self-referential
 *      (meaning that something points back to the start,
 *      such as a=b and then b=a).
 *
 * Side effects:
 *      None
 *
 *------------------------------------------------------------
 */
static int
up_check(dbPtr, varPtr, start)
     SdbFile                *dbPtr;
     SdbVariable            *varPtr;
     SdbVariable            *start;
{

    SdbDependentList       *up;
    int                     result = OK;

    /* 
     * Ascend the dependency tree checking for a reference
     * back to start.
     */
    up = varPtr->up;
    while (up) {
        SdbVariable            *v;

        if ((v = SdbFound(dbPtr->varPtr, up->id))) {
            if (v == start)
                result = CYCLING;
            else if (up_check(dbPtr, v, start) == CYCLING)
                result = CYCLING;
        }
        up = up->next;
    }
    return result;
}


/*
 * Some helper macros for the optimization routines.
 */

#define ltree  (tree->left=SdbOptimize(dbPtr, tree->left))
#define rtree  (tree->right=SdbOptimize(dbPtr, tree->right))
#define trval  tree->content.value
#define ltrval (tree->left)->content.value
#define rtrval (tree->right)->content.value

/*
 *------------------------------------------------------------
 *
 * SdbOptimize --
 *
 * scan the parse tree and attempt to reduce constant expressions. Build  
 * lists into the appropriate structures, lookup variables by name,
 * check their types, and replace them with symbolic links. 
 * Non-volatile functions with constant arguments are evaluated. 
 *
 * Results:
 *      A pointer to the (possibly) reduced tree.
 *
 * Side effects:
 *      May added to the result string if errors are detected
 *      during reduction (such as a divide by constant zero,
 *      or divide by a undefined variable.)
 *
 *------------------------------------------------------------
 */
SdbNode                *
SdbOptimize(dbPtr, tree)
     SdbFile                *dbPtr;
     SdbNode                *tree;

{
    if (tree == 0) {
        fprintf(stderr, "Tried to reduce an empty node\n");
        exit(1);
    }
    switch (tree->type) {

    case lookupCode:
    case lookupNumber:
    case constValue:
    case constCode:
        break;

    case charFnct:
        if (rtree->type == constValue) {
            tree->type = constValue;
            trval = (tree->content.fnctptr) ((tree->right)->content.code);
            tree->right = SdbFreeNode(tree->right);
        }
        break;

    case formatFnct:{
            SdbNode                *next;
	    int j=0;

            ltree;
            next = tree->right;
            while (next) {
	        if (j==10) {
		  fprintf(stderr,"too many arguments to format function\n");
		  exit(1);
		}
                next->left = SdbOptimize(dbPtr, next->left);
                next = next->right;
		j++;
            }
            break;
        }

    case simplexFnct:
        if (rtree->type == constValue) {
            tree->type = constValue;
	    trval = (tree->content.fnctptr) (rtrval);
            tree->right = SdbFreeNode(tree->right);
        }
        break;

    case duplexFnct:
        rtree;
        ltree;
        if ((tree->left)->type == constValue && 
	        (tree->right)->type == constValue) {
            tree->type = constValue;
            trval = (tree->content.fnctptr) (ltrval, rtrval);
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
        }
        break;

    case hsgFnct:
        if (rtree->type == constCode) {
            char                   *temp = (tree->right)->content.code;

            (tree->right)->content.list.n = 1;
            (tree->right)->content.list.pntr = 
	      (char **) ckalloc(sizeof(char *));

            (tree->right)->content.list.pntr[0] = temp;
            (tree->right)->type = hsgCodeList;
        }
        break;

    case hsgCode:{
            int                     new;
            SdbVariable            
	        *varPtr = SdbFound(dbPtr->varPtr, tree->content.code);

            if (!varPtr || !(varPtr->flag & SDB_CODE)) {
                tree->content.code = GetIndices(dbPtr->htbl, ".", &new);
                tree->type = constCode;
            } else if (varPtr->flag & SDB_CONSTANT) {
	      /*
	       * If the variable being referenced is a constant that
	       * can be altered by an assignment in an expression, then
	       * defer folding it into this expression.  If it is not
	       * assigned by other expressions, it is safe to fold it.
	       */
	        if (varPtr->flag & SDB_MUTABLE) {
		  tree->content.varPtr = varPtr;
		  tree->type = hsgCodeIndex;
		} else {
                   tree->content.code = varPtr->u.code;
		   tree->type = constCode;
		 }
            } else if (varPtr->flag & (SDB_EXPRESSION | SDB_LOOKUP)) {
                if (varPtr->u.expr[0]->type == constCode) {
                    tree->content.code = varPtr->u.expr[0]->content.code;
                    tree->type = constCode;
                } else {
                    tree->content.varPtr = varPtr;
                    tree->type = charExpr;
                }
            } else if (varPtr->flag & SDB_ALLOCATED) {
                tree->content.varPtr = varPtr;
                tree->type = hsgCodeIndex;
            }
        }
        break;

    case hsgNumber:{
            SdbVariable            
	        *varPtr = SdbFound(dbPtr->varPtr, tree->content.code);

            if (!varPtr || !(varPtr->flag & SDB_NUMBER)) {
                tree->content.value = 0;
                tree->type = constValue;
            } else if (varPtr->flag & SDB_CONSTANT) {
	      /*
	       * If the variable being referenced is a constant that
	       * can be altered by an assignment in an expression, then
	       * defer folding it into this expression.  If it is not
	       * assigned by other expressions, it is safe to fold it.
	       */
	        if (varPtr->flag & SDB_MUTABLE) {
		  tree->content.varPtr = varPtr;
		  tree->type = hsgNumberIndex;
		} else {
                   tree->content.value = varPtr->u.value;
		   tree->type = constValue;
		 }
            } else if (varPtr->flag & (SDB_EXPRESSION | SDB_LOOKUP)) {
	        /* 
		 * If the expression or lookup yields a constant, then
		 * fold it in. Otherwise just add a refernce and evaulate
		 * the other expression at runtime.
		 */
                if (varPtr->u.expr[0]->type == constValue) {
                    tree->content.value = varPtr->u.expr[0]->content.value;
                    tree->type = constValue;
                } else {
                    tree->content.varPtr = varPtr;
                    tree->type = numExpr;
                }
            } else if (varPtr->flag & SDB_ALLOCATED) {
                tree->content.varPtr = varPtr;
                tree->type = hsgNumberIndex;
            }
        }
        break;

    case unaryOp:
        if (rtree->type == constValue) {
            tree->content.value = evaluate(dbPtr, 0, tree);
            tree->type = constValue;
            tree->right = SdbFreeNode(tree->right);
        }
        break;

    case binaryOp:
	ltree;
	rtree;
        if (tree->right->type == constValue && tree->content.op == opDiv) {
	  /* check for divide by zero, eliminate from expression */
	    if ((tree->right)->content.value == 0.0) {
	      SdbNode *temp=tree->left;
	      if (dbPtr->interp->result && *dbPtr->interp->result)
		Tcl_AppendResult(dbPtr->interp,"\n",NULL);
	      Tcl_AppendResult(dbPtr->interp,
		     "warning: division by zero eliminated from expression",NULL);
	      tree->right = SdbFreeNode(tree->right);
	      ckfree(tree);
	      return(temp);
	    } else if ((tree->right)->content.value == 1.0) {
	      SdbNode *temp=tree->left;
	      tree->right = SdbFreeNode(tree->right);
	      ckfree(tree);
	      return(temp);
	    }
	    else {
	      tree->content.op = opMul;
	      (tree->right)->content.value = 
		1.0 / ((tree->right)->content.value);
	    }
	    break;
        }
	/* check for multiply by zero, reduce to zero */
        if (tree->content.op == opMul &&
	    ((tree->left->type == constValue && 
	      tree->left->content.value == 0.0) ||
	     (tree->right->type == constValue && 
	      tree->right->content.value == 0.0))) {
	  tree->content.value = 0.0;
	  tree->type = constValue;
	  tree->right = SdbFreeNode(tree->right);
	  tree->left = SdbFreeNode(tree->left);
	  break;
	}
	
	/* 1 * value == value */
        if (tree->content.op == opMul &&
	    ((tree->left->type == constValue && 
	      tree->left->content.value == 1.0))) {
	  SdbNode *temp=tree->right;
	  tree->left = SdbFreeNode(tree->left);
	  ckfree(tree);
	  return(temp);
	}

	/* value * 1 == value */
        if (tree->content.op == opMul &&
	    ((tree->right->type == constValue && 
	      tree->right->content.value == 1.0))) {
	  SdbNode *temp=tree->left;
	  tree->right = SdbFreeNode(tree->right);
	  ckfree(tree);
	  return(temp);
	}

	/* check for addition or subtraction of zero */
	if (((tree->content.op == opSum) || (tree->content.op == opSub)) &&
	    (tree->right->type == constValue && 
	     tree->right->content.value == 0.0)) {
	  SdbNode *temp=tree->left;
	  tree->right=SdbFreeNode(tree->right);
	  ckfree(tree);
	  return(temp);
	}

        if ((tree->content.op == opSum) && 
	    (tree->left->type == constValue && 
	     tree->left->content.value == 0.0)) {
	  SdbNode *temp=tree->right;
	  tree->left=SdbFreeNode(tree->left);
	  ckfree(tree);
	  return(temp);
	}

        if ((tree->content.op == opSub) && 
	    (tree->left->type == constValue && 
	     tree->left->content.value == 0.0)) {
	  tree->left=SdbFreeNode(tree->left);
	  tree->type = unaryOp;
	  tree->content.op = opNeg;
	  return(tree);
	}

	/* check for both arguments constant, reduce */
        if (tree->left->type == constValue && 
	          (tree->right)->type == constValue) {
            tree->content.value = evaluate(dbPtr, 0, tree);
            tree->type = constValue;
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
        }
        break;

    case booleanOp:
        rtree;
        ltree;
        if ((tree->left)->type == constValue && 
	        (tree->right)->type == constValue) {
            tree->content.value = evaluate(dbPtr, 0, tree);
            tree->type = constValue;
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
        } else if (tree->content.op == opAnd &&
                    ((((tree->right)->type == constValue) &&
                                    ((tree->right)->content.value == 0)) ||
                            (((tree->left)->type == constValue) &&
                                ((tree->left)->content.value == 0)))) {
            tree->type = constValue;
            tree->content.value = 0;
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
        } else if (tree->content.op == opOr &&
                    ((((tree->right)->type == constValue) &&
                                    ((tree->right)->content.value != 0)) ||
                            (((tree->left)->type == constValue) &&
                                ((tree->left)->content.value != 0)))) {
            tree->type = constValue;
            tree->content.value = 1;
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
        }
        break;

    case chrChrFnct:
    case numChrFnct:
    case chrNumFnct:
    case numNumFnct:{
            SdbCat                 *e = tree->content.classList;

            tree->left = ltree;
            while (e) {
                if (e->expr)
                    e->expr = SdbOptimize(dbPtr, e->expr);
                e->code = SdbOptimize(dbPtr, e->code);
                e = e->next;
                /* this could be souped up to further resolve if all expression
                 * are constant */
            }
        }
        break;

    case stringCompare:{
            ltree;
            rtree;
            if (tree->left->type == constCode && 
		    tree->right->type == constCode) {
                tree->content.value = evaluate(dbPtr, 0, tree);
                tree->type = constValue;
                tree->right = SdbFreeNode(tree->right);
                tree->left = SdbFreeNode(tree->left);
            }
        }
        break;

    case stringFind:{
            if (ltree->type == constCode) {
                tree->content.value = evaluate(dbPtr, 0, tree);
                tree->type = constValue;
                tree->left = SdbFreeNode(tree->left);
                tree->right = SdbFreeNode(tree->right);
            }
        }
        break;

    case randomFnct:{
            Tcl_HashEntry          *entryPtr;

            tree->right = rtree;

            if (tree->right->type == constCode) {
                entryPtr = Tcl_FindHashEntry(dbPtr->randomList, 
					     tree->right->content.code);
                if (entryPtr)
                    tree->content.randomPtr =
                            (SdbRandom *) Tcl_GetHashValue(entryPtr);
                else
                    tree->content.randomPtr =
                            SdbCreateRandom(dbPtr, tree->right->content.code);
                break;
            } else {
                fprintf(stderr, "non-constant random sequence name");
                exit(1);
                break;
            }
        }

    case charSequence:
    case numSequence: {
            if (tree->left)
              ltree;
            if (tree->right)
	      rtree;
            break;
        }

    case assignChar:{
      if (tree->left->type == hsgCode) {
	SdbVariable            
	  *varPtr = SdbFound(dbPtr->varPtr, tree->left->content.code);
	if (varPtr && (varPtr->flag & SDB_CODE)) 
	  varPtr->flag |= SDB_MUTABLE;
      }
      if (ltree->type != hsgCodeIndex) {
	    int new;
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
            tree->type = constCode;
            tree->content.code = GetIndices(dbPtr->htbl, ".", &new);
	    dbPtr->errors |= SDB_ASSIGNEXP;
	    return (tree);
	  }
      rtree;
      break;
    }


    case assignNum:{
      if (tree->left->type == hsgNumber) {
	SdbVariable            
	  *varPtr = SdbFound(dbPtr->varPtr, tree->left->content.code);
	if (varPtr && (varPtr->flag & SDB_NUMBER)) 
	  varPtr->flag |= SDB_MUTABLE;
      }
      if (ltree->type != hsgNumberIndex) {
            tree->right = SdbFreeNode(tree->right);
            tree->left = SdbFreeNode(tree->left);
            tree->type = constValue;
            tree->content.value = 1;
	    dbPtr->errors |= SDB_ASSIGNEXP;
	    return (tree);
	  }
      rtree;
      break;
    }

    case curveFnct:{

            Tcl_HashEntry          *entryPtr;
            SdbYieldSet            *yield_set;
            SdbYieldHandle         *yield_data = tree->content.yieldHandle;
            int                     i;

            yield_data->yieldSet = NULL;
            yield_data->xIndex = -1;
            yield_data->yIndex = -1;
            yield_data->cache = NULL;

	    /* examine the arguments that were set up by the
	     * parser. */

	    /* The name of the yield set */
            tree->left->left->left = 
	      SdbOptimize(dbPtr, tree->left->left->left);

	    /* the name of the x-axis, y-axis, and independent variable */ 
            tree->left->right = SdbOptimize(dbPtr, tree->left->right);
            tree->right->left = SdbOptimize(dbPtr, tree->right->left);
            tree->right->right = SdbOptimize(dbPtr, tree->right->right);

            if (tree->left->left->left->type == constCode) {
                entryPtr = Tcl_FindHashEntry(dbPtr->yieldList,
                        tree->left->left->left->content.code);
                if (entryPtr) {
                    yield_set = yield_data->yieldSet =
                            (SdbYieldSet *) entryPtr->clientData;

		    /* while were at it, scan the data dictionary
		       and build up a list of pointers to the 
		       variables that this curve is indexed on */
                    for (i = 0; i < yield_set->numCats; i++)
                        yield_set->list[i] = SdbFound(dbPtr->varPtr, 
                               yield_set->cats[i]);

		    /* now check for variable name aliases.  If
		       a list of aliases is present then scan down
		       the list and build up and variable alias list. */
		    if (tree->left->left->right) {
		      SdbNode *top=tree->left->left->right;

		      /* make sure there is space for the alias pointers */
		      if (!yield_data->list) {
			yield_data->list=
			  (SdbVariable**)ckalloc(yield_set->numCats*
						 sizeof(SdbVariable*));
		      }

		      /* start off with a copy of the old list */
		      for (i = 0; i < yield_set->numCats; i++)
                        yield_data->list[i] = yield_set->list[i];

		      /* now scan through the alias list and
			 make substitutions.  There are two possible
			 options here.  The alias list is a single
			 assignment, or it is a sequence of assignments.*/
		      if (top->type == assignChar) {
			for (i = 0; i < yield_set->numCats; i++) {
			  if (!strcmp(yield_set->cats[i],top->left->content.code)) {
			    yield_data->list[i] = SdbFound(dbPtr->varPtr,
							   top->right->content.code);
			  }
			}
		      }

		      /* code for the assignment sequence */
		      else {
			while (top) {
			  for (i = 0; i < yield_set->numCats; i++) {
			    if (!strcmp(yield_set->cats[i],top->right->left->content.code)) {
			      yield_data->list[i] = SdbFound(dbPtr->varPtr,
							     top->right->right->content.code);
			      break;
			    }
			  }
			  top = top->left;
			}
		      }
		    }
		      

                    if (tree->left->right->type == constCode) {
                        for (i = 0; i < yield_set->numValues; i++) {
                            if (!strcmp(tree->left->right->content.code,
                                            yield_set->values[i])) {
                                yield_data->xIndex = i;
                                break;
                            }
                        }
                    }
                    if (tree->right->left->type == constCode) {
                        for (i = 0; i < yield_set->numValues; i++) {
                            if (!strcmp(tree->right->left->content.code,
                                            yield_set->values[i])) {
                                yield_data->yIndex = i;
                                break;
                            }
                        }
                    }
                } else {
                    /* no entryPtr!  There is no table for this
		     * entry, so delete the function. */
		  dbPtr->errors |= SDB_MISSYIELD;
		  ckfree(tree->content.yieldHandle);
		  tree->right = SdbFreeNode(tree->right);
		  tree->left = SdbFreeNode(tree->left);
		  tree->type = constValue;
		  tree->content.value = 0.0;
		  return(tree);
                }
	      }
	    else {
	      fprintf(stderr,"non-constant yield set name");
	      exit(1);
	    }
	  }
        break;

    case ifNode:{
            tree->right = rtree;
            tree->left = ltree;
            if ((tree->left)->type == constValue && ltrval != 0) {
                SdbNode                *temp = tree->right;

                tree->left = SdbFreeNode(tree->left);
                ckfree(tree);
                return (temp);
            }
        }
        break;

    default:
        fprintf(stderr, "Unknown node in SdbOptimize\n");
        exit(1);
    }

    return tree;
}

/*
 * Some macros to help with the evaluation functions
 */
#define lval evaluate(dbPtr,n,tree->left)
#define rval evaluate(dbPtr,n,tree->right)
#define lcval charEval(dbPtr,n,tree->left)
#define rcval charEval(dbPtr,n,tree->right)

/*
 *------------------------------------------------------------
 *
 * charEval --
 *
 *      Run-time evaluation engine for character expressions.
 *
 * Results:
 *      The string resulting from the evauation of an expression..
 *
 * Side effects:
 *      May added to the result string if errors are detected
 *      during evaluation.
 *
 *------------------------------------------------------------
 */
static char                   *
charEval(dbPtr, n, tree)
     SdbFile                *dbPtr;
     int                     n;
     SdbNode                *tree;

/* evaluate a parse tree */
{
    int                     new;

    if (tree == 0) {
        fprintf(stderr, "Null parse tree\n");
        exit(1);
    }
    switch (tree->type) {
    case constCode:
        return (tree->content.code);

    case charExpr:
        return (charEval(dbPtr, n, (tree->content.varPtr)->u.expr[0]));

    case numChrFnct:{
            double                  first = evaluate(dbPtr, n, tree->left);
            SdbCat                 *e = tree->content.classList;

            while (e) {
                double                  v;
                int                     i;

                if (e->expr) {
                    v = evaluate(dbPtr, n, e->expr);
                    switch (e->relation) {
                    case opGe:
                        i = first >= v;
                        break;
                    case opGt:
                        i = first > v;
                        break;
                    case opLe:
                        i = first <= v;
                        break;
                    case opLt:
                        i = first <= v;
                        break;
                    case opEq:
                        i = first == v;
                        break;
                    case opNe:
                        i = first != v;
                        break;
                    case opNop:
                        i = 1;
                        break;
                    default:{
                            fprintf(stderr, "Invalid test in numChrFnct\n");
                            exit(1);
                        }
		      }
                } else
                    i = 1;
                if (i)
                    return (charEval(dbPtr, n, e->code));
                e = e->next;
            }
            return (GetIndices(dbPtr->htbl, ".", &new));
        }

    case chrChrFnct:{
            char                   *first = charEval(dbPtr, n, tree->left);
            SdbCat                 *e = tree->content.classList;

            while (e) {
                char                   *v;
                int                     i;

                if (e->expr) {
                    v = charEval(dbPtr, n, e->expr);
		    i = strcmp(first, v);
		    switch (e->relation) {
		    case opEq:
		      i =  (i == 0);
		      break;
		    case opLt:
		      i = (i < 0);
		      break;
		    case opLe:
		      i = (i <= 0);
		      break;
		    case opGt:
		      i = (i > 0);
		      break;
		    case opGe:
		      i = (i <= 0);
		      break;
		    case opNe:
		      i = (i != 0);
		      break;
                    case opNop:
                        i = 1;
                        break;
                    default:{
                            fprintf(stderr, "Invalid test in numChrFnct\n");
                            exit(1);
			  }
                    }
                } else
                    i = 1;
                if (i) {
		  if (e->code)
                    return (charEval(dbPtr, n, e->code));
		  else 
		    return (GetIndices(dbPtr->htbl, ".", &new));
		}
                e = e->next;
            }
            return (GetIndices(dbPtr->htbl, ".", &new));
        }

    case assignChar:{
            return (SdbPutCode(dbPtr, tree->left->content.varPtr, 
			       charEval(dbPtr, n, tree->right), n));
        }

    case charSequence:{
            if (tree->left) {
                if (tree->left->type == numSequence)
                    evaluate(dbPtr, n, tree->left);
                else
                    charEval(dbPtr, n, tree->left);
            }
            return charEval(dbPtr, n, tree->right);
        }

    case hsgCodeIndex:
        if ((tree->content.varPtr)->flag & SDB_ALLOCATED)
            return (tree->content.varPtr)->u.ivalue[n];
        else if ((tree->content.varPtr)->flag & SDB_CONSTANT)
            return (tree->content.varPtr)->u.code;
        else {
            fprintf(stderr, "bad variable type");
            exit(1);
        }
        break;


    case lookupCode:
        return (SdbCodeLookup(dbPtr, tree->content.luHandlePtr,
                        (tree->right)->content.integer, n));

    case formatFnct:{
            struct {
                union {
                    char                   *code;
                    double                  value;
                } u;
            } ptr[10];
            int                     j;
            char                    temp[256],
                                   *formatSpec;
            SdbNode                *next = tree->right;

            /* first evaluate the format specifier */
            formatSpec = lcval;

            /* load up the argument list */
            j = 0;
            while (next) {
                if (j == 10) {
                    fprintf(stderr, "too many argmuents to format command");
                    exit(1);
                }
                if (next->type >= constCode)
                    ptr[j].u.code = charEval(dbPtr, n, next->left);
                else
                    ptr[j].u.value = evaluate(dbPtr, n, next->left);
                next = next->right;
                j++;
            }

            /* evaluate the function */
	    switch (j) {
	    case 0:
	      sprintf(temp, formatSpec);
	      break;
	    case 1:
	      sprintf(temp, formatSpec, ptr[0].u.value);
	      break;
	    case 2:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value);
	      break;
	      break;
	    case 3:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value);
	      break;
	    case 4:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value);
	      break;
	    case 5:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value, ptr[4].u.value);
	      break;
	    case 6:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value, ptr[4].u.value,
		    ptr[5].u.value);
	      break;
	    case 7:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value, ptr[4].u.value,
		    ptr[5].u.value, ptr[6].u.value);
	      break;
	    case 8:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value, ptr[4].u.value,
		    ptr[5].u.value, ptr[6].u.value, ptr[7].u.value);
	      break;
	    case 9:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value, ptr[4].u.value,
		    ptr[5].u.value, ptr[6].u.value, ptr[7].u.value,
		    ptr[8].u.value);
	      break;
	    case 10:
	      sprintf(temp, formatSpec, ptr[0].u.value, ptr[1].u.value,
		    ptr[2].u.value, ptr[3].u.value, ptr[4].u.value,
		    ptr[5].u.value, ptr[6].u.value, ptr[7].u.value,
		    ptr[8].u.value,ptr[9].u.value);
	    }
            /* store the result in the global hash table.  This is
             * regretable but necessary.  Otherwise, the transient value
             * would get lost on the stack. */
            return (GetIndices(dbPtr->htbl, temp, &new));
        }

    default:
        fprintf(stderr, "Unknown operation in char_eval\n");
        exit(1);
    }
    return (GetIndices(dbPtr->htbl, ".", &new));

}


/*
 *------------------------------------------------------------
 *
 * evaluate --
 *
 *      Run-time evaluation engine for numeric expressions.
 *
 * Results:
 *      A double resulting from the evauation of an expression..
 *
 * Side effects:
 *      May added to the result string if errors are detected
 *      during evaluation.
 *
 *------------------------------------------------------------
 */
static double
evaluate(dbPtr, n, tree)
     SdbFile                *dbPtr;
     int                     n;
     SdbNode                *tree;

{

    if (tree == 0) {
        fprintf(stderr, "Null parse tree\n");
        exit(1);
    }
    switch (tree->type) {
    case constValue:
        return tree->content.value;

    case charFnct:
        return (tree->content.fnctptr) (rcval);

    case simplexFnct:
        return (tree->content.fnctptr) (rval);

    case duplexFnct:
        return (tree->content.fnctptr) (lval, rval);

    case unaryOp:
        switch (tree->content.op) {
        case opNeg:
            return -rval;
        case opNot:
            return !rval;

        default:
            fprintf(stderr, "Unknown unary op in tree\n");
            exit(1);
        }
        break;

    case numExpr:
        return (evaluate(dbPtr, n, (tree->content.varPtr)->u.expr[0]));

    case assignNum:{
            return (SdbPutNumber(dbPtr, tree->left->content.varPtr,
                            evaluate(dbPtr, n, tree->right), n));
        }

    case numSequence:{
            if (tree->left) {
                if (tree->left->type == numSequence)
                    evaluate(dbPtr, n, tree->left);
                else
                    charEval(dbPtr, n, tree->left);
            }
            return evaluate(dbPtr, n, tree->right);
            break;
        }

    case numNumFnct:{
            double                  first = evaluate(dbPtr, n, tree->left);
            SdbCat                 *e = tree->content.classList;

            while (e) {
                double                  v;
                int                     i;

                if (e->expr) {
                    v = evaluate(dbPtr, n, e->expr);
                    switch (e->relation) {
                    case opGe:
                        i = first >= v;
                        break;
                    case opGt:
                        i = first > v;
                        break;
                    case opLe:
                        i = first <= v;
                        break;
                    case opLt:
                        i = first <= v;
                        break;
                    case opEq:
                        i = first == v;
                        break;
                    case opNe:
                        i = first != v;
                        break;
                    case opNop:
                        i = 1;
                        break;
                    default:{
                            fprintf(stderr, "Invalid test in numNumFnct\n");
                            exit(1);
                        }
                    }
                } else
                    i = 1;
                if (i)
                    return (evaluate(dbPtr, n, e->code));
                e = e->next;
            }
            return (0.0);
        }

    case chrNumFnct:{
            char                   *first = charEval(dbPtr, n, tree->left);
            SdbCat                 *e = tree->content.classList;

            while (e) {
                char                   *v;
                int                     i;

                if (e->expr) {
                    v = charEval(dbPtr, n, e->expr);
		    i = strcmp(first, v);
		    switch (e->relation) {
		    case opEq:
		      i =  (i == 0);
		      break;
		    case opLt:
		      i = (i < 0);
		      break;
		    case opLe:
		      i = (i <= 0);
		      break;
		    case opGt:
		      i = (i > 0);
		      break;
		    case opGe:
		      i = (i <= 0);
		      break;
		    case opNe:
		      i = (i != 0);
		      break;
                    case opNop:
                        i = 1;
                        break;
                    default:{
                            fprintf(stderr, "Invalid test in numNumFnct\n");
                            exit(1);
			  }
                    }
                } else
                    i = 1;
                if (i) {
		  if (e->code)
                    return (evaluate(dbPtr, n, e->code));
		  else 
		    return (0.0);
		}
                e = e->next;
            }
            return (0.0);
        }

    case binaryOp:
        switch (tree->content.op) {
        case opSum:
            return lval + rval;
        case opSub:
            return lval - rval;
        case opMul:
            return lval * rval;
        case opDiv:{
                double                  a = lval,
                                        b = rval;

                if (b == 0) {
		  dbPtr->errors |= SDB_DIVZERO;
		  return 0;
		}
		return a / b;
	      }

        case opEq:
            return (lval == rval);
        case opLe:
            return (lval <= rval);
        case opLt:
            return (lval < rval);
        case opGe:
            return (lval >= rval);
        case opGt:
            return (lval > rval);
        case opNe:
            return (lval != rval);

        default:
            fprintf(stderr, "Unknown binary op code in evaluate\n");
            exit(1);
	  }
        break;

    case hsgFnct:{
            if ((tree->right)->type == hsgCodeList)
                return ((tree->content.fnctptr) 
			(n, (tree->right)->content.list));
            else {
                char                   *i;  /* Untested, this may have bugs! */
                SdbList                 list;

		list.n=1;
		list.pntr=&i;
                i = rcval;
                return ((tree->content.fnctptr) (n, list));
            }
        }
        break;

    case curveFnct:{

            SdbYieldSet            *yield_set;
            SdbYieldHandle         *yield_data = tree->content.yieldHandle;
            int                     x_index = -1,
                                    y_index = -1;

            char                   *x_var;
            char                   *y_var;
            int                     i;

            if (yield_data->yieldSet)
                yield_set = yield_data->yieldSet;
            else {
	      fprintf(stderr,"tried to evaluate a missing yield set\n");
	      exit(1);;
            }

            if (yield_data->xIndex >= 0)
                x_index = yield_data->xIndex;
            else {
                x_var = charEval(dbPtr, n, tree->left->right);
                for (i = 0; i < yield_set->numValues; i++) {
                    if (!strcmp(x_var, yield_set->values[i])) {
                        x_index = i;
                        break;
                    }
                }
                if (x_index < 0) {
		    dbPtr->errors |= SDB_MISSYIELD;
                    return 0.0;
		  }
            }

            if (yield_data->yIndex >= 0)
                y_index = yield_data->yIndex;
            else {
                y_var = charEval(dbPtr, n, tree->right->left);
                for (i = 0; i < yield_set->numValues; i++) {
                    if (!strcmp(y_var, yield_set->values[i])) {
                        y_index = i;
                        break;
                    }
                }
                if (y_index < 0) {
		    dbPtr->errors |= SDB_MISSYIELD;
                    return 0.0;
		  }
            }
            return (SdbYieldLookup(dbPtr, yield_data, n, tree->right->right));
        }

    case hsgNumberIndex:{
            if ((tree->content.varPtr)->flag & SDB_ALLOCATED)
                return (tree->content.varPtr)->u.dvalue[n];
            else if ((tree->content.varPtr)->flag & SDB_CONSTANT)
                return (tree->content.varPtr)->u.value;
            else {
                fprintf(stderr, "bad variable type");
                exit(1);
            }
            break;
        }

    case randomFnct:{
            double                  value;

            if (!tree->content.randomPtr) {
                fprintf(stderr, "random state not initialized!");
                exit(1);
            }
            setstate((tree->content.randomPtr)->state);
            value = random();
            value /= (unsigned long)2147483647;
	    if (tree->content.randomPtr->type == normal) {
	      float p=value, x, mean=0.0, sd=1.0, bounds;
	      int which=2, status;
	      p=(p<1.0e-7)?1.0e-7:((p>1.0-1.0e-7)?1.0-1.0e-7:p);
	      cdfnor_(&which,&p,&x,&mean,&sd,&status,&bounds);
	      if (status != 0) {
		fprintf(stderr,"error in normal distribution function\n");
		exit(1);
	      }
	      value=x;
	    }
            return (value);
        }

    case booleanOp:{
            int                     i;

            switch (tree->content.op) {
            case opAnd:
                i = lval && rval;
                break;
            case opOr:
                i = lval || rval;
                break;
            default:
                fprintf(stderr, "Invalid boolean op code in evalaute\n");
                exit(1);
            }
            return i;
        }
        break;

    case ifNode:{
            if (lval)
                return (rval);
            else
                return (0);
        }
        break;

    case stringCompare:{
            int                     i;

            i = strcmp(lcval, rcval);
            switch (tree->content.op) {
            case opEq:
                return (i == 0);
            case opLt:
                return (i < 0);
            case opLe:
                return (i <= 0);
            case opGt:
                return (i > 0);
            case opGe:
                return (i <= 0);
            case opNe:
                return (i != 0);
            default:
                fprintf(stderr, "Invalid string compare code in evaluate\n");
                exit(1);
                break;
            }
        }
        break;

    case stringFind:{
            int                     i;
            char                   *st = lcval;
            SdbNode                *rt = tree->right;

            if (rt->type != hsgCodeList) {
                fprintf(stderr, "Wrong argument to string find\n");
                exit(1);
            }
            for (i = 0; i < rt->content.list.n; i++)
                if (strcmp(st, rt->content.list.pntr[i]))
                    return TRUE;
            return FALSE;
        }
        break;

    case lookupNumber:
        return (SdbNumberLookup(dbPtr, tree->content.luHandlePtr,
                        (tree->right)->content.integer, n));

    default:
        fprintf(stderr, "Unknown operation in evaluate\n");  /* something wrong */
        exit(1);
    }
    return 0;            
}

/*
 * Here are the entry points for the evaluation functions.
 * They set flags values once and then call the recursive
 * evaluation routines.
 */
double
SdbEvaluate(dbPtr, n, tree)
     SdbFile                *dbPtr;
     int                     n;
     SdbNode                *tree;
{
  double value;

  value = evaluate(dbPtr, n, tree);
  return(value);
}

char                   *
SdbCharEval(dbPtr, n, tree)
     SdbFile                *dbPtr;
     int                     n;
     SdbNode                *tree;
{
  char *code;

  code=charEval(dbPtr, n, tree);
  return(code);
}

/*
 * Two functions to help manage math errors.  Division by zero
 * is prevented, and a flag is set.  These functions clear the
 * flag and reports the most recent errors.
 */

static void
reset_error_status (dbPtr)
      SdbFile                *dbPtr;
{
  dbPtr->errors = 0;
}

int
SdbErrorCheck(dbPtr)
      SdbFile                *dbPtr;
{
  if (dbPtr->errors & SDB_DIVZERO) {
    if (dbPtr->interp->result && *dbPtr->interp->result)
      Tcl_AppendResult(dbPtr->interp,"\n",NULL);
    Tcl_AppendResult(dbPtr->interp,
		     "division by zero altered to return zero",NULL);
  }
  if (dbPtr->errors & SDB_MISSYIELD) {
    if (dbPtr->interp->result && *dbPtr->interp->result)
      Tcl_AppendResult(dbPtr->interp,"\n",NULL);
    Tcl_AppendResult(dbPtr->interp,
		     "yield curve not found",NULL);
  }
  if (dbPtr->errors & SDB_ASSIGNEXP) {
    if (dbPtr->interp->result && *dbPtr->interp->result)
      Tcl_AppendResult(dbPtr->interp,"\n",NULL);
    Tcl_AppendResult(dbPtr->interp,
		     "invalid assignment, expression eliminated",NULL);
  }
  reset_error_status(dbPtr);
  return TCL_OK;
}

/*
 *------------------------------------------------------------
 *
 * SdbDefine --
 *
 *      Define a variable as an expression.  Calls the parser 
 *      and then the optimizer to process the expression string.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      If successful, creates a new variable of the appropriate
 *      type.
 *
 *------------------------------------------------------------
 */
int
SdbDefine(interp, dbPtr, name, string)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *name;
     char                   *string;
{
    SdbNode                *e;
    int                     s_type = 0;
    SdbVariable            *vars,
                           *v;

    if ((vars = SdbFound(dbPtr->varPtr, name))) {
        Tcl_AppendResult(interp, "variable \"", name, 
			 "\" already exists, cannot redefine", NULL);
        return TCL_ERROR;
    }
    if ((e = SdbParse(interp, dbPtr, string)) == NULL) {
        return TCL_ERROR;
    }

    /*
     * Figure out what type of expression has been entered.
     */
    if (e->type >= constCode)
        s_type = SDB_CODE;
    else
        s_type = SDB_NUMBER;
    s_type |= SDB_OUTPUT;

    if ((e->type == constValue) || (e->type == constCode))
        s_type |= SDB_CONSTANT;
    else
        s_type |= SDB_EXPRESSION;

    /*
     * Insert it into a slot.
     */
    vars = SdbInsert(dbPtr, name, s_type, 0);

    if (s_type & SDB_CONSTANT) {
        if (s_type & SDB_NUMBER) {
            vars->u.value = e->content.value;
        } else {
            vars->u.code = e->content.code;
        }
        e=SdbFreeNode(e);
    } else {
        vars->u.expr[1] = e;
    }

    /* 
     * Rebuild the dependency tree (a list of what variables each 
     * expression depends on).
     */
    for (v = dbPtr->varPtr; v; v = v->next)
        if (v->flag & SDB_EXPRESSION) {
            v->down = down_add(dbPtr, v->u.expr[1], v->down,
                    v->id);
	} else if (v->flag & SDB_LOOKUP) {
            v->down = down_add(dbPtr, v->u.expr[0], v->down,
			       v->id);
        }
    /* 
     * Scan the dependency tree and check for cycling, (e.g. 
     * where definition of 'a = b', and also 'b = a').
     * These would cause infinite loops in the evaluation machinery. 
     */
    if (up_check(dbPtr, vars, vars) == CYCLING) {
        interp->result = "Cycling detected, expresion rejected";
        SdbDelete(interp, dbPtr, vars->id, 0);
        return TCL_ERROR;
    }

    /*
     * re-optimize the code in all expressions.  This is essential
     * because some constants may be mutated by assignements.
     */
    SdbReOptimize(dbPtr);

    Tcl_AppendResult(interp, vars->id, (char *) NULL);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbCalulate --
 *
 *      Apply an expression to all selected records.  Calls the
 *      parser, then the optimizer.  If the expression looks ok
 *      it loops over all selected records and calls the
 *      evaluator.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      May change the values of other variables.  Appends the
 *      sum of numeric evaluation or the final character evaluation
 *      to the result string.
 *
 *------------------------------------------------------------
 */
int
SdbCalculate(interp, dbPtr, string)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *string;
{

    int                     type,
                            n,
                            caused_mutation,
                            rec;
    _bit_type              *selectPtr = dbPtr->select;
    char                   *code=".",
                            temp[32];
    double                  value=0.0;
    SdbNode                *e;
    SdbVariable            *varPtr;

    /*
     * Parse and optimize the expression
     */
    if ((e = SdbParse(interp, dbPtr, string)) == NULL) {
        return TCL_ERROR;
    }
    e = SdbOptimize(dbPtr, e);
    /*
     * This expression may mutate existing constant variables.
     * Freshen up the code in all expressions depending on mutated
     * constants to take care of this.
     */
    if ((caused_mutation=does_assignment(e)))
      for (varPtr=dbPtr->varPtr; varPtr; varPtr=varPtr->next) {
	if (varPtr->flag & SDB_MUTABLE)
	  SdbFreshenUp(dbPtr, varPtr);
      }

    if (e->type >= constCode)
        type = SDB_CODE;
    else
        type = SDB_NUMBER;

    /* 
     * loop through and apply to all stands. 
     */
    for (n = 0; n < dbPtr->numRec; n++) {
      rec = dbPtr->base[n];
      if (bit_test(selectPtr, rec)) {
	dbPtr->stamp++;
	if (type == SDB_CODE) 
	  code = SdbCharEval(dbPtr, rec, e);
	else 
	  value += SdbEvaluate(dbPtr, rec, e);
      }
    }

    dbPtr->stamp++;
    e=SdbFreeNode(e);
    /*
     * Now that the transient expression has been deleted,
     * reoptimize the rest of the expression tree.  
     */
    if (caused_mutation)
      SdbReOptimize(dbPtr);

    /* 
     * prepare the result string
     */
    if (type == SDB_CODE)
        Tcl_AppendResult(interp, code, NULL);
    else {
        sprintf(temp, "%-.15g", value);
        Tcl_AppendResult(interp, temp, NULL);
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbDelete --
 *
 *      Delete a variable.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      Refreshes all expressions.  If other expression depend on
 *      the variable that is being deleted then errors may be 
 *      generated.
 *
 *------------------------------------------------------------
 */
int
SdbDelete(interp, dbPtr, string, fast)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     CONST char             *string;  /* the name of the variable to delete */
     int                     fast;    /* optimization flag:  0 means refresh
				       * the dictionary after deleting this
				       * variable; 1 means don't bother. */
{
    SdbVariable            *vars,
                          **prev;
    SdbDependentList       *next,
                           *save;

    /*
     * Search for the variable of interest
     */
    for (prev = &dbPtr->varPtr, vars = dbPtr->varPtr;
            vars;
            prev = &vars->next, vars = vars->next) {
        if ((strcmp(vars->id, string) == 0))
            break;
    }

    if (!vars) {
        Tcl_AppendResult(interp, "no such variable: \"", string, "\"", NULL);
        return TCL_ERROR;
    }

    /* 
     * If the database is indexed on this field, set the pointer to NULL.
     */
    if (vars->flag & SDB_INDEXED)
        dbPtr->indexPtr = NULL;

    /*
     * Steps to take for deleting various kinds of variables
     *
     * For allocated varaibles, free the storage.
     */
    if (vars->flag & SDB_ALLOCATED) {
        if (vars->flag & SDB_CODE) {
            ckfree(vars->u.ivalue);
        } else {
            ckfree(vars->u.dvalue);
        }

    /* 
     * For expressions, free the expression 
     */
    } else if (vars->flag & SDB_EXPRESSION) {
        free_expression(dbPtr, vars);

    /* 
     * For lookups, free the lookup handle.  Delete the handle and the
     * join tree if the reference count decrements to zero.
     */
    } else if (vars->flag & SDB_LOOKUP) {
        SdbNode                *ptr = vars->u.expr[0];
        SdbLookupHandle        *luHandlePtr = ptr->content.luHandlePtr;
        SdbLookup              *luPtr = luHandlePtr->luPtr;

        ptr=SdbFreeNode(ptr);
        if ((--luHandlePtr->referenceCount) == 0) {
            Tcl_HashEntry          *entryPtr;

            ckfree(luHandlePtr->index);
            entryPtr = Tcl_FindHashEntry(dbPtr->joinList, luHandlePtr->name);
            Tcl_DeleteHashEntry(entryPtr);
            if (!luHandlePtr->remote) {
                SdbDelJoinTree(luPtr->head);
                luPtr->head = NULL;
            }
            ckfree(luHandlePtr);
        }
        if ((--luPtr->referenceCount) == 0) {
            ckfree(luPtr->indexNames);
            ckfree(luPtr->codeNames);
            ckfree(luPtr->valueNames);
            ckfree(luPtr);
        }
    }
    if (vars->format)
        ckfree(vars->format);

    /* get rid of the up and down pointers */
    for (next = vars->up; next; next = save) {
        save = next->next;
        ckfree(next);
    }
    free_down_up(dbPtr, vars->id, vars->down);

    /* remove this element from the linked list */
    *prev = vars->next;

    ckfree(vars);

    /* Kill the trace */
    Tcl_UnsetVar2(interp, dbPtr->dbname, (char *) string, TCL_GLOBAL_ONLY);

    /* tidy up the structures */
    if (!fast) {
      SdbReOptimize(dbPtr);
      SdbRebuildOutput(dbPtr);
    }

    Tcl_SetResult(interp, (char*)string, TCL_VOLATILE);
    return TCL_OK;

}


/*
 *------------------------------------------------------------
 *
 * SdbFreshenUp --
 *
 *      Climb the dependency tree and Re-optimize the code for all 
 *      defined expressions and lookup tables.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Refreshes all expressions and lookup variables.  
 *
 *------------------------------------------------------------
 */
void
SdbFreshenUp(dbPtr, varPtr)
     SdbFile *dbPtr;
     SdbVariable *varPtr;
{
  SdbDependentList       *next=varPtr->up;
  
  /*
   * First freshen the code for this expression
   */
  if (varPtr->flag & SDB_EXPRESSION) {
    /* 
     * Refresh the compiled code just in case
     * an important link has been removed or added.  
     */
    varPtr->u.expr[0] = SdbFreeNode(varPtr->u.expr[0]);
    varPtr->u.expr[0] = tree_copy(varPtr->u.expr[1]);
    varPtr->u.expr[0] = SdbOptimize(dbPtr, varPtr->u.expr[0]);
  } else if ((varPtr->flag & SDB_LOOKUP)) {
    /*
     * In the case of a lookup table, refresh the pointers
     * to the index variables.
     */
    SdbLookupHandle        
      *luHandlePtr = (varPtr->u.expr[0])->content.luHandlePtr;
    SdbLookup              *luPtr = luHandlePtr->luPtr;
    int i;

    for (i = 0; i < luPtr->numIndex; i++) {
      luHandlePtr->index[i] = 
	SdbFound(dbPtr->varPtr, luPtr->indexNames[i]);
    }
  }

  /*
   * Scan up the dependency tree and freshen the code of 
   * expressions that depend on this variable.  This might
   * result in expressions being optimized over and over again.
   * I can't think of a safe way to do it otherwise.
   */
  while(next) {
    SdbVariable            *v;
 
    v = SdbFound(dbPtr->varPtr, next->id);
    if (v)
      SdbFreshenUp(dbPtr,v);
    next=next->next;
  }
}


/*
 *------------------------------------------------------------
 *
 * SdbReOptimize --
 *
 *      Re-optimize the code for all defined expressions.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Rebuilds the dependency trees.
 *
 *------------------------------------------------------------
 */
void
SdbReOptimize(dbPtr)
     SdbFile                *dbPtr;
{

    SdbVariable            *varPtr;

    /*
     * Turn off the MUTABLE bit.  This will be checked by
     * next down_add procedure.
     */
    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next)
	varPtr->flag &= ~SDB_MUTABLE;
    /*
     * Loop through all variables and refresh the dependency lists.
     * This must be done because a variable that is deleted then
     * reinserted will have an empty 'up' pointer.  The only
     * way to reconstruct it is to do the following loop.
     */

    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next) {
      if (varPtr->flag & SDB_EXPRESSION) {
	varPtr->down = down_add(dbPtr, varPtr->u.expr[1], varPtr->down,
				varPtr->id);
      } else if (varPtr->flag & SDB_LOOKUP) {
	varPtr->down = down_add(dbPtr, varPtr->u.expr[0], varPtr->down,
				varPtr->id);
      }
    } 

    /* 
     * now scan the variable list again and optimize 
     */
    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next) {
      SdbFreshenUp(dbPtr,varPtr);
    }
  }


/*
 *------------------------------------------------------------
 *
 * SdbSelect --
 *
 *      Manipulate the selection bitmap.  Toggle bits on and off
 *      to determine the selected set of database records.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      May create or delete selection bitmaps, or alter the
 *      selected records in a set.
 *
 *------------------------------------------------------------
 */
int
SdbSelect(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    _bit_type              *selectPtr = dbPtr->select;
    int                     n = dbPtr->numRec;
    int                     rec,
                            i,
                            caused_mutation=0,
                            j,
                            result = TCL_OK;
    char                    temp[20];
    SdbNode                *e=NULL;;

    if (argc == 0) {
        Tcl_AppendResult(interp, "too few arguments", NULL);
        return TCL_ERROR;
    }
    if (argc > 3) {
        Tcl_AppendResult(interp, "too many arguments", NULL);
        return TCL_ERROR;
    }
    if (!strcmp("first", argv[0]) ||
            !strcmp("last", argv[0]) ||
            !strcmp("next", argv[0]) ||
            !strcmp("pop", argv[0]) ||
            !strcmp("push", argv[0])) {
        if (argc >= 2) {
            Tcl_AppendResult(interp, "too many arguments", NULL);
            return TCL_ERROR;
        } else if (!strcmp("first", argv[0])) {
            for (i = 0; i < dbPtr->numRec && !bit_test(selectPtr, 
						       dbPtr->base[i]); i++) ;
            if (i < dbPtr->numRec) {
                sprintf(temp, "%d", i);
                Tcl_AppendResult(interp, temp, NULL);
                return TCL_OK;
            } else {
                Tcl_AppendResult(interp, "no records selected", NULL);
                return TCL_ERROR;
            }
        } else if (!strcmp("next", argv[0])) {
            for (i = dbPtr->position + 1; 
		 i < dbPtr->numRec && !bit_test(selectPtr, dbPtr->base[i]); 
		 i++) ;
            if (i < dbPtr->numRec) {
                sprintf(temp, "%d", i);
                Tcl_AppendResult(interp, temp, NULL);
                return TCL_OK;
            } else {
                Tcl_AppendResult(interp, "no records remaining", NULL);
                return TCL_ERROR;
            }
        } else if (!strcmp("last", argv[0])) {
            for (i = dbPtr->numRec - 1; 
		 i >= 0 && !bit_test(selectPtr, dbPtr->base[i]); 
		 i--) ;
            if (i >= 0) {
                sprintf(temp, "%d", i);
                Tcl_AppendResult(interp, temp, NULL);
                return TCL_OK;
            } else {
                Tcl_AppendResult(interp, "no records selected", NULL);
                return TCL_ERROR;
            }
        } else if (!strcmp("pop", argv[0])) {
            SdbStack               *top;

            if (!dbPtr->stack) {
                Tcl_AppendResult(interp, "stack empty", NULL);
                return TCL_ERROR;
            }
            top = dbPtr->stack;
            ckfree(dbPtr->select);
            dbPtr->stack = top->next;
            dbPtr->select = top->select;
            ckfree(top);
            j = 0;
            for (i = 0; i < n; i++)
                if (bit_test(dbPtr->select, dbPtr->base[i]))
                    j++;
            dbPtr->numSelect = j;
            sprintf(temp, "%d", j);
            Tcl_AppendResult(interp, temp, NULL);
            return TCL_OK;
        } else if (!strcmp("push", argv[0])) {
            SdbStack               *top = dbPtr->stack;
            dbPtr->stack = (SdbStack *) ckalloc(sizeof(SdbStack));

            (dbPtr->stack)->next = top;
            (dbPtr->stack)->select = dbPtr->select;
            dbPtr->select = (_bit_type *) ckalloc(_bit_size(dbPtr->numRec) *
                    sizeof(_bit_type));

            bcopy((char *) (dbPtr->stack)->select,
                    (char *) dbPtr->select,
                    (_bit_size(dbPtr->numRec) * 4));
            sprintf(temp, "%d", dbPtr->numSelect);
            Tcl_AppendResult(interp, temp, NULL);
            return TCL_OK;
        }
    }

    /* 
     * check to see if a specific record is being queried...
     */
    if (argc == 2) {
      if (!strcmp(argv[0], "status")) {
	if (Tcl_GetInt(interp,argv[1],&rec) != TCL_OK)
	  return TCL_ERROR;
	if (rec < 0 || rec > n) {
	  Tcl_AppendResult(interp, "record out of range: ", argv[1],NULL);
	  return TCL_ERROR;
	}
	if (bit_test(selectPtr, dbPtr->base[rec]))
	  Tcl_SetResult(interp,"1",TCL_STATIC);
	else
	  Tcl_SetResult(interp,"0",TCL_STATIC);
	return TCL_OK;
      }
    }

    /* 
     * check to see if a specific record is being toggled...
     */
    if (argc == 3) {
        if (!strcmp(argv[1], "record")) {
            if (Tcl_GetInt(interp,argv[2],&rec) != TCL_OK)
	      return TCL_ERROR;
        } else {
            Tcl_AppendResult(interp, "invalid keyword \"", argv[1],
                    "\", expecting \"record\"", NULL);
            return TCL_ERROR;
        }
        if (rec < 0 || rec > n) {
            Tcl_AppendResult(interp, "record out of range: ", argv[2]);
            return TCL_ERROR;
        }
    }
    /* 
     * ...or parse the expression (if present) 
     */
    else if (argc == 2) {
        if ((e = SdbParse(interp, dbPtr, argv[1])) == NULL) {
            return TCL_ERROR;
        }
        e = SdbOptimize(dbPtr, e);
	caused_mutation=does_assignment(e);
        if (e->type >= constCode) {
            Tcl_AppendResult(interp,
                    "character expression not allowed in select statement",
                    NULL);
            result = TCL_ERROR;
            goto clean_up;
        }

	/*
	 * This expression may mutate existing constant variables.
	 * Freshen up the code in all expressions depending on mutated
	 * constants to take care of this.
	 */
	if (caused_mutation) {
	  SdbVariable *varPtr;
	  for (varPtr=dbPtr->varPtr; varPtr; varPtr=varPtr->next) {
	    if (varPtr->flag & SDB_MUTABLE)
	      SdbFreshenUp(dbPtr, varPtr);
	  }
	}
	
    }


    /*
     * All arguments have been read and now process the
     * set of records.  If this command has only one argument,
     * that means apply the operation to the entire set of
     * records.  With 2 arguments, apply the operation to the
     * set of records that match the expression.  With 3
     * arguments, apply the expression to only the single record
     * that was specified on the command line.
     */

    /*
     * AND - apply to records that are already selected
     */
    if (!strcmp("and", argv[0])) {
        j = dbPtr->numSelect;
        if (argc == 2) {
            for (i = 0; i < n; i++) {
                dbPtr->stamp++;
                if (bit_test(selectPtr, dbPtr->base[i]) &&
                        !SdbEvaluate(dbPtr, dbPtr->base[i], e)) {
                    bit_clear(selectPtr, dbPtr->base[i]);
                    j--;
                }
            }
        }
        dbPtr->numSelect = j;
        sprintf(temp, "%d", j);
        Tcl_AppendResult(interp, temp, NULL);
    } 

    /*
     * NOT - toggle selection bits (off becomes on, on becomes off).
     */
    else if (!strcmp("not", argv[0])) {
        if (argc == 1) {
            for (i = 0; i < n; i++)
                bit_not(selectPtr, dbPtr->base[i]);
            dbPtr->numSelect = n - dbPtr->numSelect;
        } else if (argc == 2) {
            j = 0;
            for (i = 0; i < n; i++) {
                dbPtr->stamp++;
                if (SdbEvaluate(dbPtr, i, e))
                    bit_not(selectPtr, dbPtr->base[i]);
                if (bit_test(selectPtr, dbPtr->base[i]))
                    j++;
            }
            dbPtr->numSelect = j;
        } else if (argc == 3) {
            bit_not(selectPtr, rec);
            if (bit_test(selectPtr, rec))
                dbPtr->numSelect++;
            else
                dbPtr->numSelect--;
        }
        sprintf(temp, "%d", dbPtr->numSelect);
        Tcl_AppendResult(interp, temp, NULL);
    } 

    /*
     * SET - like an 'or', turn bits on
     */
    else if (!strcmp("set", argv[0])) {
        if (argc == 1) {
            bit_setall(selectPtr, n);
            dbPtr->numSelect = n;
        } else if (argc == 2) {
            j = dbPtr->numSelect;
            for (i = 0; i < n; i++) {
                dbPtr->stamp++;
                if (!bit_test(selectPtr, dbPtr->base[i]) &&
                        SdbEvaluate(dbPtr, dbPtr->base[i], e)) {
                    bit_set(selectPtr, dbPtr->base[i]);
                    j++;
                }
            }
            dbPtr->numSelect = j;
        } else if (argc == 3) {
            if (!bit_test(selectPtr, rec)) {
                bit_set(selectPtr, rec);
                dbPtr->numSelect++;
            }
        }
        sprintf(temp, "%d", dbPtr->numSelect);
        Tcl_AppendResult(interp, temp, NULL);
    } 

    /*
     * UNSET - turn bits off
     */
    else if (!strcmp("unset", argv[0])) {
        if (argc == 1) {
            bit_clearall(selectPtr, n);
            dbPtr->numSelect = 0;
        } else if (argc == 2) {
            j = dbPtr->numSelect;
            for (i = 0; i < n; i++) {
                dbPtr->stamp++;
                if (bit_test(selectPtr, dbPtr->base[i]) && 
		    SdbEvaluate(dbPtr, dbPtr->base[i], e)) {
		  bit_clear(selectPtr, dbPtr->base[i]);
		  j--;
                }
            }
            dbPtr->numSelect = j;
        } else if (argc == 3) {
            if (bit_test(selectPtr, rec)) {
                bit_clear(selectPtr, rec);
                dbPtr->numSelect--;
            }
        }
        sprintf(temp, "%d", dbPtr->numSelect);
        Tcl_AppendResult(interp, temp, NULL);
    } 

    /*
     * otherwise, this is an invalid command.
     */
    else {
        Tcl_AppendResult(interp, "invalid option \"", argv[0],
                "\", should be one of:  and, current, first, last, not, pop, push, set or unset", NULL);
        result = TCL_ERROR;
    }

  clean_up:
    if (e)
        e = SdbFreeNode(e);
    if (caused_mutation)
      SdbReOptimize(dbPtr);
    return result;

  }

