/* 
 * sdbJoin.c --
 *
 *  This file contains the relational join command and yield
 *  curve interpolation routines.
 *
 * Copyright (c) 1995 Tom Moore.  All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all copies of this software.
 * 
 */

#include "sdbInt.h"
#include "tree.h"
#include <unistd.h>

#define DELIMITOR "="
#define SEP " \n"  
#define END "\n"


/* GLOBAL VARIABLES */
static int              added_values;  /* number of added values on the end 
					  of the rule record */
static int              rclass;        /* number of rule classes */
static int              added_codes;   /* number of additional rule classes */
static int              lineno;        /* line number counter */
static int              currentDepth;  /* accessed by recursive tree routine */
static char           **classList;     /* " */
static Tcl_Interp      *currentInterp; /* " */



/* Local procedures */
static int process _ANSI_ARGS_((Tcl_Interp *interp, SdbTree *head, 
				SdbTerminal *terminal, int match_col[], 
				char **listV, Tcl_HashTable *htbl, char **path,
				int k, int l));
static int SdbOld _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
			       SdbTextFile *file, int *match_col, 
			       char **match_id, SdbLookup *luPtr));
static int SdbNew _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
			       SdbTextFile *file, int *match_col, 
			       char **match_id));
static int SdbJoinLoad _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    int argc, char **argv));
static int              place _ANSI_ARGS_((SdbTree *head, char **path, int n, 
			      SdbTerminal *terminal));
static int              comp_label1 _ANSI_ARGS_((char *label, SdbTree *node));
static int              comp_label2 _ANSI_ARGS_((SdbTree *to, SdbTree *from));
static SdbTree         *make_node _ANSI_ARGS_((char* label));
static int rec_remove _ANSI_ARGS_((SdbTree *head, char **path, int n));
static int data_proc _ANSI_ARGS_((Tcl_Interp *interp, Tcl_HashTable *htbl, 
				  SdbTree *head, SdbTerminal *terminal, 
				  int *match_col, char **path, 
				  SdbTextFile *file));
static void           **SdbLuSearch _ANSI_ARGS_((SdbFile *dbPtr, 
						 SdbTree *head, 
				    SdbVariable **list, int limit, int current,
				    int rec));
static void           **SdbLuDirectSearch _ANSI_ARGS_((SdbFile *dbPtr, 
						 SdbTree *head, 
				    char **list, int limit, 
						       int current));
static void process_point _ANSI_ARGS_((SdbYieldSet *yield_set, char **path, 
				       double values[]));
static double SdbInterpolate _ANSI_ARGS_((SdbCurve *point, int index1, 
					  int index2, double x));
static int SdbGetCurve _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    int argc, char **argv));
static int SdbDefineCurve _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    int argc, char **argv));
static int list_classes _ANSI_ARGS_((SdbTree *node));
static int SdbLoadYields _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    int argc, char **argv));
static int SdbCreateSets _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    int argc, char **argv));
static void SdbDelYieldTree _ANSI_ARGS_((SdbTree *node));
static void trav _ANSI_ARGS_((Tcl_Interp *interp, FILE *output, tree **ppr_tree,
			      SdbLookup *luPtr, char **path, int i));
static void saveHelper _ANSI_ARGS_((Tcl_Interp *interp, FILE *output, 
				     SdbTree *head,
			      SdbLookup *luPtr, char **path, int i));
static int SdbJoinSave _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    SdbLookup *luPtr, int argc, char **argv));
static int SdbJoinInsert _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    SdbLookupHandle *luHandlePtr, int argc, 
				      char **argv));
static int SdbJoinDeleteRec _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    SdbLookup *luPtr, int argc, char **argv));
static int SdbJoinInfo _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    SdbLookup *luPtr, int argc, char **argv));
static int SdbJoinConnect _ANSI_ARGS_((Tcl_Interp *interp, SdbFile *dbPtr, 
				    int argc, char **argv));




/*
 *------------------------------------------------------------
 *
 * SdbJoinCmd --
 *
 *      The main dispatching command for yield tables.
 *      This routine calls the other join table management routines
 *      to do the work.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      Many, depending on the arguments.
 *
 *------------------------------------------------------------
 */
int
SdbJoinCmd(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    Tcl_HashEntry          *entryPtr;
    SdbLookup              *luPtr;
    SdbLookupHandle        *luHandlePtr;

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    /* 
     * try to load up a new table (or possibly reload an existing table).
     * Pass all arguments into this routine.  It will check for an
     * existing table and create one it required. 
     */
    if (!strcmp("load", argv[0]))
        return (SdbJoinLoad(interp, dbPtr, argc - 1, argv + 1));
    if (!strcmp("connect", argv[0]))
        return (SdbJoinConnect(interp, dbPtr, argc - 1, argv + 1));

    /* 
     * the following functions operate on existing tables, 
     * so make sure it exists first 
     */
    entryPtr = Tcl_FindHashEntry(dbPtr->joinList, argv[1]);
    if (!entryPtr) {
        Tcl_AppendResult(interp, "join table \"", argv[1], 
			 "\" not found", NULL);
        return TCL_ERROR;
    }
    luHandlePtr = (SdbLookupHandle *) Tcl_GetHashValue(entryPtr);
    luPtr = luHandlePtr->luPtr;
    if (!luPtr->head) {
        Tcl_AppendResult(interp, "remote join table \"", argv[1],
                "\" has been deleted", NULL);
        return TCL_ERROR;
    }
    if (!strcmp("save", argv[0]))
        return (SdbJoinSave(interp, dbPtr, luPtr, argc - 2, argv + 2));
    else if (!strcmp("insert", argv[0]))
        return (SdbJoinInsert(interp, dbPtr, luHandlePtr, argc - 2, argv + 2));
    else if (!strcmp("info", argv[0]))
        return (SdbJoinInfo(interp, dbPtr, luPtr, argc - 2, argv + 2));
    else if (!strcmp("delete", argv[0]))
        return (SdbJoinDeleteRec(interp, dbPtr, luPtr, argc - 2, argv + 2));
    else {
        Tcl_AppendResult(interp, "bad option \"", argv[0],
                "\": should be connect, delete, info, insert, load or save", 
			 NULL);
        return TCL_ERROR;
    }
}


/*
 *------------------------------------------------------------
 *
 * SdbCurves --
 *
 *      Main routine for managing yield curve sets.
 *
 * Results:
 *      Returns a stadard TCL error code.
 *
 * Side effects:
 *      Calls the other yield curve management routines
 *      to do the work.
 *
 *------------------------------------------------------------
 */
int
SdbCurves(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    if (argc == 0) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    if (!strcmp(argv[0], "load")) {
        if (argc != 3) {
            Tcl_AppendResult(interp, "wrong number of arguments", NULL);
            return TCL_ERROR;
        }
        return (SdbLoadYields(interp, dbPtr, argc - 1, &argv[1]));
    } else if (!strcmp(argv[0], "fields")) {
        Tcl_HashEntry          *entryPtr;
        SdbYieldSet            *yield_set;
        int                     i;

        if (argc != 2) {
            Tcl_AppendResult(interp, "wrong number of arguments", NULL);
            return TCL_ERROR;
        }
        if (!(entryPtr = Tcl_FindHashEntry(dbPtr->yieldList, argv[1]))) {
            Tcl_AppendResult(interp, "curve set \"", argv[1], 
			     "\" not found", NULL);
            return TCL_ERROR;
        }
        yield_set = (SdbYieldSet *) Tcl_GetHashValue(entryPtr);
        Tcl_AppendResult(interp, "{", NULL);
        for (i = 0; i < yield_set->numCats; i++)
            Tcl_AppendElement(interp, yield_set->cats[i]);
        Tcl_AppendResult(interp, "} {", NULL);
        for (i = 0; i < yield_set->numValues; i++)
            Tcl_AppendElement(interp, yield_set->values[i]);
        Tcl_AppendResult(interp, "}", NULL);
        return TCL_OK;
    } else if (!strcmp(argv[0], "list")) {
        Tcl_HashEntry          *entryPtr;
        SdbYieldSet            *yield_set;

        if (argc != 2) {
            Tcl_AppendResult(interp, "wrong number of arguments", NULL);
            return TCL_ERROR;
        }
        if (!(entryPtr = Tcl_FindHashEntry(dbPtr->yieldList, argv[1]))) {
            Tcl_AppendResult(interp, "curve set \"", argv[1], 
			     "\" not found", NULL);
            return TCL_ERROR;
        }
        yield_set = (SdbYieldSet *) Tcl_GetHashValue(entryPtr);
        currentInterp = interp;
        currentDepth = 0;
        classList = (char **) ckalloc(yield_set->numCats * sizeof(char *));

        tree_trav((tree **) & yield_set->tree->tree, list_classes);
        ckfree(classList);
        return TCL_OK;
    } else if (!strcmp(argv[0], "sets")) {
        Tcl_HashEntry          *entryPtr;
        Tcl_HashSearch          search;

        if (argc != 1) {
            Tcl_AppendResult(interp, "wrong number of arguments", NULL);
            return TCL_ERROR;
        }
        for (entryPtr = Tcl_FirstHashEntry(dbPtr->yieldList, &search);
                entryPtr != NULL;
                entryPtr = Tcl_NextHashEntry(&search))
            Tcl_AppendElement(interp, Tcl_GetHashKey(dbPtr->yieldList, 
						     entryPtr));
    } else if (!strcmp(argv[0], "get")) {
        return (SdbGetCurve(interp, dbPtr, argc - 1, &argv[1]));
    } else if (!strcmp(argv[0], "create")) {
        return (SdbCreateSets(interp, dbPtr, argc - 1, &argv[1]));
    } else if (!strcmp(argv[0], "define")) {
        return (SdbDefineCurve(interp, dbPtr, argc - 1, &argv[1]));
    } else {
        Tcl_AppendResult(interp, "bad option \"", argv[0],
                "\": should be create, define, fields, get, list, load or sets",
                (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * data_proc --
 *
 *      reads a record from the join file and loads it into 
 *      the terminal structure
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
data_proc(interp, htbl, head, terminal, match_col, path, file)
     Tcl_Interp             *interp;
     Tcl_HashTable          *htbl;
     SdbTree                *head;
     SdbTerminal            *terminal;
     int                    *match_col;
     char                  **path;
     SdbTextFile            *file;
{
    double                 *value = terminal->value;
    int                     i, j;

    for (j=0, i = file->split; i < file->n; j++, i++) {
        value[j] = 1.0;
	if (!sscanf(file->list[i], "%lf", &value[j])) {
                    Tcl_AppendResult(interp,
                            "invalid value in rule file: ", file->list[i],
                            NULL);
                    value[j] = 0.0;
                    return TCL_ERROR;
                }
      }
    if (process(interp, head, terminal, match_col, file->list,
                    htbl, path, 0, 0) == TCL_ERROR) {
        return TCL_ERROR;
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbOld --
 *
 *      Parses the field names from a join table, and makes
 *      sure that they match the previous table.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
SdbOld(interp, dbPtr, file, match_col, match_id, luPtr)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     SdbTextFile            *file;
     char                  **match_id;
     int                    *match_col;
     SdbLookup              *luPtr;
{

    int                    codeC, valueC;
    Tcl_HashTable          *htbl = dbPtr->htbl;
    int                     i,
                            j,
                            new;

    codeC = file->split;
    valueC = file->n - file->split;

    /*
     * check the number of fields in the file.  It must be exactly the
     * same as in the orignial file.
     */
    if ((luPtr->numIndex + luPtr->numCodes) != codeC ||
            luPtr->numVals != valueC) {
        Tcl_AppendResult(interp, "wrong number of fields in file", NULL);
        return TCL_ERROR;
    }
    for (i = 0; i < rclass + added_values; i++)
        match_id[i] = NULL;

    /* 
     * search for matching index fields 
     */
    for (i = 0; i < luPtr->numIndex; i++) {
        for (j = 0; j < file->split; j++) {
            if (strcmp(luPtr->indexNames[i], file->field[j]) == 0) {
                match_id[j] = 
		  Tcl_CreateHashEntry(htbl, file->field[j], &new)->key.string;
                match_col[j] = j;
                break;
            }
        }
        if (!match_id[i]) {
	  Tcl_AppendResult(interp, "required index field not found in file", 
			   NULL);
	  return TCL_ERROR;
	}
    }

    /* 
     * search for matching added code fields 
     */
    for (i = 0; i < luPtr->numCodes; i++) {
        for (j = 0; j < file->split; j++) {
            if (strcmp(luPtr->codeNames[i], file->field[j]) == 0) {
                match_id[j] = 
		  Tcl_CreateHashEntry(htbl, file->field[j], &new)->key.string;
                match_col[j] = -j - 1;
                break;
            }
        }
        if (!match_id[i]) {
	  Tcl_AppendResult(interp, "required code field not found in file", 
			   NULL);
	  return TCL_ERROR;
	}
    }

    /* 
     * search for matching added value fields 
     */
    for (i = 0; i < luPtr->numVals; i++) {
        for (j = file->split; j < file->n; j++) {
            if (strcmp(luPtr->valueNames[i], file->field[j]) == 0) {
                match_id[i + rclass] = 
		  Tcl_CreateHashEntry(htbl, file->field[j], &new)->key.string;
                break;
            }
        }
        if (!match_id[i + rclass]) {
	  Tcl_AppendResult(interp, "required value field not found in file", 
			   NULL);
	  return TCL_ERROR;
	}
    }

    /* 
     * so far, so good.  It looks like this is a good replacement file 
     */
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbNew --
 *
 *      Parses the field names from a join table, and sets
 *      up a new structure.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
SdbNew(interp, dbPtr, file, match_col, match_id)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     SdbTextFile            *file;
     char                  **match_id;
     int                    *match_col;
{
    Tcl_HashTable          *htbl = dbPtr->htbl;
    SdbVariable            *vars;
    int                     j, i, new;

    j = added_codes = 0;

    /* 
     * loop through test the field names against existing variable
     * names.  
     */
    for (i=0; i<file->split; i++) {
      if ((vars = SdbFound(dbPtr->varPtr, file->field[i]))) {
	if (vars->flag & SDB_NUMBER) {
	  Tcl_AppendResult(interp,
			   "join column is numeric \"", file->field[i], "\"",
			   NULL);
	  return TCL_ERROR;
	}
	match_id[j] = (char *) vars->id;
	match_col[j] = j;
      } else {
	match_col[j] = -j - 1;
	match_id[j] = Tcl_CreateHashEntry(htbl, file->field[i], 
					  &new)->key.string;
	added_codes++;
      }
      j++;
    }
    rclass = j;

    /*
     * Make sure that at least one index column is defined.
     */
    if ((added_codes - j) == 0) {
      Tcl_AppendResult(interp,
		       "join table does not contain matching label",
		       NULL);
      return TCL_ERROR;
    }

    /* 
     * Now loop through and pull out value labels 
     */
    for (i=file->split; i<file->n; i++) {
        if (SdbFound(dbPtr->varPtr, file->field[i])) {
            Tcl_AppendResult(interp, "attempt to redefine variable \"",
                    file->field[i], "\"", NULL);
            return TCL_ERROR;
        }
        match_id[j] = Tcl_CreateHashEntry(htbl, file->field[i], 
					  &new)->key.string;
        j++;
    }
    added_values = j - rclass;
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbJoinLoad --
 *
 *      Loads a join table from a file.  This can be a new table
 *      or a replacement for an existing table.  If a replacement
 *      then the field names must be identical.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      Creates a new table structure.  May add new varaibles to the
 *      database.
 *
 *------------------------------------------------------------
 */
static int
SdbJoinLoad(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    SdbTextFile            *file;
    SdbTree                *head;
    int                     i,
                            k,
                            l,
                            result = TCL_OK,
                            new;
    double                 *value;
    char                  **match_id = NULL,
                          **path = NULL,
                          **codes;
    int                    *match_col = NULL;
    Tcl_HashTable          *htbl = dbPtr->htbl;
    SdbTerminal            *terminal = NULL;
    int                     listC;
    Tcl_HashEntry          *entryPtr;
    SdbLookup              *luPtr = NULL;
    
    /* 
     * check the arguments 
     */
    if (argc != 2) {
        Tcl_AppendResult(interp, 
		"wrong number of arguments: load luName fileName",
                NULL);
        return TCL_ERROR;
    }

    /* 
     * check the hash table for the lu table name 
     */
    entryPtr = Tcl_FindHashEntry(dbPtr->joinList, argv[0]);
    if (entryPtr)
        luPtr = ((SdbLookupHandle *) Tcl_GetHashValue(entryPtr))->luPtr;

    /*
     * grab a new top level pointer for this tree
     */
    head = make_node((char *) NULL);

    /*
     * Try opening the file using the TextFile routines
     */
    lineno = 0;
    file = SdbTextOpen(interp,argv[1],8192);
    if (!file) {
        Tcl_AppendResult(interp, "could not open file: ", argv[1], NULL);
        return TCL_ERROR;
    }

    /*
     * Parse out the column names
     */

    /* if the luPtr already exists, then we must be reloading
     * a file into an existing table.  First, check to see if
     * the columns match.  If they do, then delete the old 
     * table and proceed with reloading. */
    if (luPtr) {
      added_codes = luPtr->numCodes;
      added_values = luPtr->numVals;
      rclass = added_codes + luPtr->numIndex;
      listC = luPtr->numIndex + luPtr->numCodes + luPtr->numVals + 1;
      match_col = (int *) ckalloc(listC * sizeof(int));
      match_id = (char **) ckalloc(listC * sizeof(char *));
      
      if (SdbOld(interp, dbPtr, file, match_col, match_id, luPtr) 
	  != TCL_OK) {
	result = TCL_ERROR;
	goto clean_up;
      }

      /* if the VAR statement parsed ok, then get rid of the
       * old tree and prepare for a new one. */
      SdbDelJoinTree(luPtr->head);
      luPtr->head = NULL;
    } 

    /* If the luPtr did not exist previously, then
     * we are loading a table for the first time.
     * Extra steps are required to determine which
     * of the fields are for indexing or adding. 
     */
    else {
      listC = file->n;
      match_col = (int *) ckalloc(listC * sizeof(int));
      match_id = (char **) ckalloc(listC * sizeof(char *));
      if (SdbNew(interp, dbPtr, file, match_col, match_id) != TCL_OK) {
	result = TCL_ERROR;
	goto clean_up;
      }
    }

    /*
     * Now make space for structures that hold data record contents
     */
    value = (double *) ckalloc(added_values * sizeof(double));
    codes = (char **) ckalloc(added_codes * sizeof(char *));
    terminal = (SdbTerminal *) ckalloc(sizeof(SdbTerminal));
    
    terminal->codes = codes;
    terminal->value = value;
    path = (char **) ckalloc(rclass * sizeof(char *));
  
    /*
     * Now process all of the data records in the join file.
     */
    while (SdbTextGet(file)) {
      if (data_proc(interp, htbl, head, terminal, match_col, path, file) != 
	  TCL_OK) {
	result = TCL_ERROR;
	break;
      }
    }

  clean_up:

    /* here is where we end up after reading the contents of the
     * file to be loaded.  First close the file, and then decide
     * if we should continue or abort. */

    SdbTextClose(file);
    if (result == TCL_OK) {

        Tcl_AppendResult(interp, argv[0], NULL);

        /* Continuing....  Now decide if we are reloading an existing
         * table, or creating a new table.  If reloading an existing
         * table then just update the index tree.  If loading
         * a new table then many steps have to be taken. */

        if (luPtr) {
            luPtr->head = head;
        } else {

            SdbLookupHandle        *luHandlePtr;

            /* create the new lookup table structure, first the detail... */
            luPtr = (SdbLookup *) ckalloc(sizeof(SdbLookup));
            luPtr->head = head;
            luPtr->indexNames = (char **) ckalloc((rclass - added_codes) * 
                   sizeof(char *));
            luPtr->codeNames = (char **) ckalloc(added_codes * sizeof(char *));
            luPtr->valueNames = (char **) ckalloc(added_values * 
                   sizeof(char *));
            luPtr->referenceCount = 0;
            luPtr->numCodes = added_codes;
            luPtr->numVals = added_values;

            /* then the handle information */
            luHandlePtr = (SdbLookupHandle *) ckalloc(sizeof(SdbLookupHandle));
            luHandlePtr->luPtr = luPtr;
            luHandlePtr->index =
                    (SdbVariable **) ckalloc((rclass - added_codes) * 
                    sizeof(SdbVariable *));
            luHandlePtr->cache = NULL;
            luHandlePtr->remote = NULL;
            luHandlePtr->referenceCount = 0;
            luHandlePtr->stamp = 0;
            luHandlePtr->name = GetIndices(dbPtr->htbl, argv[0], &new);

            /* add it to the catalog */
            entryPtr = Tcl_CreateHashEntry(dbPtr->joinList, argv[0], &new);
            Tcl_SetHashValue(entryPtr, luHandlePtr);

            /* now add the variables to the var structure of the current 
             * database. */
            k = l = 0;
            for (i = 0; i < rclass; i++) {
                /* first add in the new code fields */
                if (match_col[i] < 0) {
                    if (!SdbInsertLookup(dbPtr, match_id[i],
                                    SDB_LOOKUP | SDB_CODE | SDB_OUTPUT, 
					 luHandlePtr, k)) {
                        Tcl_AppendResult(interp, 
					 "failed to insert variable \"",
                                match_id[i], "\" into table", NULL);
                    } else {
                        luPtr->referenceCount++;
                        luHandlePtr->referenceCount++;
                        luPtr->codeNames[k] = match_id[i];
                    }
                    k++;
                } else {
                    luHandlePtr->index[l] = SdbFound(dbPtr->varPtr, 
						     match_id[i]);
                    luPtr->indexNames[l] = (luHandlePtr->index[l])->id;
                    l++;
                }
            }
            luPtr->numIndex = l;

            /* then add in the numeric fields */
            for (i = 0; i < added_values; i++) {
                k = i + rclass;
                if (!SdbInsertLookup(dbPtr, match_id[k],
                                SDB_LOOKUP | SDB_NUMBER | SDB_OUTPUT, 
				     luHandlePtr, i)) {
                    Tcl_AppendResult(interp, "failed to insert variable \"",
                            match_id[k], "\" into table", NULL);
                } else {
                    luPtr->referenceCount++;
                    luHandlePtr->referenceCount++;
                    luPtr->valueNames[i] = match_id[k];
                }
            }
        }
    } else {
        /* This is where we end up upon failure to load the file.
         * Make sure to get rid of the old index tree. */
        SdbDelJoinTree(head);
    }

    /* clean up temporary structures */
    if (path)
        ckfree(path);
    if (match_col)
        ckfree(match_col);
    if (match_id)
        ckfree(match_id);
    if (terminal) {
        if (terminal->codes)
            ckfree(terminal->codes);
        if (terminal->value)
            ckfree(terminal->value);
        ckfree(terminal);
    }
    /* tidy up the database */
    SdbReOptimize(dbPtr);

    return result;
    
  }


/*
 *------------------------------------------------------------
 *
 * process --
 *
 * this recursively called routine builds up the tree structure
 * that indexes the table to be joined.
 *
 * Results:
 *      Returns TCL_OK
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
process(interp, head, terminal, match_col, listV, htbl, path, k, l)
     Tcl_Interp             *interp;
     SdbTree                *head;
     SdbTerminal            *terminal;
     int                     match_col[];
     char                  **listV;
     Tcl_HashTable          *htbl;
     char                  **path;
     int                     k;
     int                     l;
{

    int                     result,
                            new;
    char                   *group,
                           *label;

    /* if k==rclass then we have recursed to the end of the list */
    if (k == rclass) {
        result = place(head, path, k - l, terminal);
    }
    /* otherwise, continue to recursively add nodes to the list */
    else {
        if (match_col[k] < 0) {
            /* if this is an added column then just add the string */
            terminal->codes[l] =
                    (Tcl_CreateHashEntry(htbl, listV[k], &new))->key.string;
            l++;
            result = process(interp, head, terminal, match_col,
                    listV, htbl, path, k + 1, l);
        } else {
            /* if this is a looup column then break the string up
             * into substrings */
            group = listV[k];
            while (group && (label = strtok(group, "/"))) {
                group = strtok(NULL, "");
                path[k - l] = 
		  (Tcl_CreateHashEntry(htbl, label, &new))->key.string;
                result = process(interp, head, terminal, match_col,
                        listV, htbl, path, k + 1, l);
            }
        }
    }

    return (result);
}


/*
 *------------------------------------------------------------
 *
 * make_node --
 *
 *      Allocates space for a new node.
 *
 * Results:
 *      Returns the address of the new node.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static SdbTree         *
make_node(label)
     char                   *label;
{
    SdbTree                *node;
    node = (SdbTree *) ckalloc(sizeof(SdbTree));

    node->label = label;
    node->wild = NULL;
    node->terminal = NULL;
    tree_init((tree **) & node->tree);
    return (node);
}


/*
 *------------------------------------------------------------
 *
 * rec_remove --
 *
 *      Prunes a single leaf from a join table tree.
 *
 * Results:
 *      Returns TCL_OK.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
rec_remove(head, path, n)
     SdbTree                *head;
     char                  **path;
     int                     n;
{
    int                     k;
    SdbTree                *ptr;

    k = 0;
    while (k < n - 1) {
        if (!strcmp(path[k], "*")) {
            if (head->wild)
                head = head->wild;
            else
                return TCL_OK;
        } else {
            if (!(ptr = (SdbTree *) tree_srch((tree **) & head->tree,
                                    comp_label1, path[k]))) {
                return TCL_OK;
            }
            head = ptr;
        }
        k++;
    }
    if (!strcmp(path[k], "*")) {
        SdbDelJoinTree(head->wild);
        head->wild = NULL;
    } else {
        ptr = make_node(path[k]);
        tree_delete((tree **) & head->tree, comp_label2, ptr, SdbDelJoinTree);
        ckfree(ptr);
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * place --
 *
 *      Inserts a leaf in a join table tree.
 *
 * Results:
 *      Returns TCL_OK
 *
 * Side effects:
 *      Creates the branches necessary to get from the root to 
 *      the leaf.
 *
 *------------------------------------------------------------
 */
static int
place(head, path, n, terminal)
     SdbTree                *head;
     char                  **path;
     int                     n;
     SdbTerminal            *terminal;
{
    int                     k;
    SdbTree                *ptr;

    k = 0;
    while (k < n) {
        if (!strcmp(path[k], "*")) {
            if (head->wild)
                ptr = head->wild;
            else {
                ptr = head->wild = make_node(path[k]);
            }
            head = ptr;
        } else {
            if (!(ptr = (SdbTree *) tree_srch((tree **) & head->tree,
                                    comp_label1, path[k]))) {
                ptr = make_node(path[k]);
                tree_add((tree **) & head->tree, comp_label2, ptr, 
			 SdbDelJoinTree);
            }
            head = ptr;
        }
        k++;
    }

    /* At this stage we should be at the end of the tree.  Insert 
     * or replace a node containing the values read off of the rule 
     * record. */
    if (!head->terminal) {
        SdbTerminal            *new;

        /* there may be an alignment problem here on some machine 
         * architectures.  Doubles often need to start on an 8 byte 
         * boundary.  'SdbTerminal' is 8 bytes long, so we should
         * be ok but watch out! */
        new = (SdbTerminal *) ckalloc(sizeof(SdbTerminal) +
                added_values * sizeof   (double) +
                added_codes * sizeof    (char *));
        new->value = (double *) ((int) new + sizeof(SdbTerminal));
        new->codes = (char **) ((int) new->value + added_values * 
				sizeof(double));

        head->terminal = new;
    }
    memcpy(head->terminal->value, terminal->value, 
	   added_values * sizeof(double));
    memcpy(head->terminal->codes, terminal->codes, 
	   added_codes * sizeof(char *));

    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * cmp_label1, cmp_label2 --
 *
 *      comparison routines needed by the AVL tree routines.
 *
 * Results:
 *      Returns an int to describe if the nodes being comapred are
 *      less, equal, or greater.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
comp_label1(label, node)
     char                   *label;
     SdbTree                *node;
{
    return (strcmp(label, node->label));
}

static int
comp_label2(to, from)
     SdbTree                *to;
     SdbTree                *from;
{
    return (strcmp(to->label, from->label));
}


/*
 *------------------------------------------------------------
 *
 * SdbDelJoinTree --
 *
 *      Delete all branches from a join tree
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Frees memory.
 *
 *------------------------------------------------------------
 */
void
SdbDelJoinTree(node)
     SdbTree                *node;
{
    if (!node)
        return;
    if (node->tree)
        tree_mung((tree **) & node->tree, SdbDelJoinTree);
    if (node->wild)
        SdbDelJoinTree(node->wild);
    if (node->terminal) {
        ckfree(node->terminal);
    }
    ckfree(node);
    return;
}


/*
 *------------------------------------------------------------
 *
 * SdbLuSearch --
 *
 *      Searches a join tree for a match.  Calls 'SdbGetCode' to
 *      get the index values for the current record.
 *
 * Results:
 *      Returns a pointer to the terminal structure for the
 *      match, or NULL if no match is found.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static void           **
SdbLuSearch(dbPtr, head, list, limit, current, rec)
     SdbFile                *dbPtr;
     SdbTree                *head;
     SdbVariable           **list;
     int                     limit;
     int                     current;
     int                     rec;
{

    void                  **result = NULL;
    SdbTree                *ptr;
    char                   *code;

    if (!head)
        return (NULL);
    if (current == limit)
        return ((void **) &head->terminal);

    if (!list[current])
        code = ".";
    else
        code = SdbGetCode(dbPtr, list[current], rec);
    if ((ptr = (SdbTree*)tree_srch((tree **) & head->tree, comp_label1, code))) {
        result = SdbLuSearch(dbPtr, ptr, list, limit, current + 1, rec);
    }
    if (!result && head->wild) {
        result = SdbLuSearch(dbPtr, head->wild, list, limit, current + 1, rec);
    }
    if (!result && head->terminal) {
        result = (void **) &head->terminal;
    }
    return result;
}


/*
 *------------------------------------------------------------
 *
 * SdbLuDirectSearch --
 *
 *      Searches a join tree for a match.  Similar to SdbLuSearch,
 *      but the index values are passed in a list.  No need to call
 *      'SdbGetCode'.
 *
 * Results:
 *      Returns a pointer to the terminal structure for the
 *      match, or NULL if no match is found.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static void           **
SdbLuDirectSearch(dbPtr, head, list, limit, current)
     SdbFile                *dbPtr;
     SdbTree                *head;
     char                  **list;
     int                     limit;
     int                     current;
{

    void                  **result = NULL;
    SdbTree                *ptr;
    char                   *code;

    if (!head)
        return (NULL);
    if (current == limit)
        return ((void **) &head->terminal);

    code = list[current];
    if ((ptr = (SdbTree*)tree_srch((tree **) & head->tree, 
	     comp_label1, code))) {
        result = SdbLuDirectSearch(dbPtr, ptr, list, limit, current + 1);
    }
    if (!result && head->wild) {
        result = SdbLuDirectSearch(dbPtr, head->wild, list, limit, current + 1);
    }
    if (!result && head->terminal) {
        result = (void **) &head->terminal;
    }
    return result;
}


/*
 *------------------------------------------------------------
 *
 * SdbCodeLookup --
 *
 *      Searches a join table for a value.  Calls SdbLuSearch to
 *      find the record, pulls out the required field, stores the
 *      record pointer in the cache.
 *
 * Results:
 *      Returns the address of the code string.
 *
 * Side effects:
 *      Updates the cache.
 *
 *------------------------------------------------------------
 */
char                   *
SdbCodeLookup(dbPtr, luHandlePtr, varIndex, rec)
     SdbFile                *dbPtr;
     SdbLookupHandle        *luHandlePtr;
     int                     varIndex;
     int                     rec;
{

    SdbLookup              *luPtr = luHandlePtr->luPtr;
    static char            *missing = ".";

    SdbTerminal           **cache;

    if (dbPtr->stamp != luHandlePtr->stamp) {
        cache = (SdbTerminal **) SdbLuSearch(dbPtr, luPtr->head,
                luHandlePtr->index,
                luPtr->numIndex, 0, rec);
        if (cache)
            luHandlePtr->cache = *cache;
        else
            luHandlePtr->cache = NULL;
        luHandlePtr->stamp = dbPtr->stamp;
    }
    if (luHandlePtr->cache)
        return ((luHandlePtr->cache)->codes[varIndex]);
    else
        return (missing);
}


/*
 *------------------------------------------------------------
 *
 * SdbPutCodeLookup --
 *
 *      Places a value in a join table.  Calls SdbLuSearch to
 *      find the record, inserts the required field, stores the
 *      record pointer in the cache.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Updates the cache.
 *
 *------------------------------------------------------------
 */
void
SdbPutCodeLookup(dbPtr, luHandlePtr, varIndex, rec, code)
     SdbFile                *dbPtr;
     SdbLookupHandle        *luHandlePtr;
     int                     varIndex;
     int                     rec;
     char                   *code;
{

    SdbLookup              *luPtr = luHandlePtr->luPtr;
    SdbTerminal           **cache;
    int                     new;

    if (dbPtr->stamp != luHandlePtr->stamp) {
        cache = (SdbTerminal **) SdbLuSearch(dbPtr, luPtr->head,
                luHandlePtr->index,
                luPtr->numIndex, 0, rec);
        if (cache)
            luHandlePtr->cache = *cache;
        else
            luHandlePtr->cache = NULL;
        luHandlePtr->stamp = dbPtr->stamp;
    }
    if (luHandlePtr->remote)
        code = GetIndices(luHandlePtr->remote, code, &new);
    if (luHandlePtr->cache)
        (luHandlePtr->cache)->codes[varIndex] = code;
}


/*
 *------------------------------------------------------------
 *
 * SdbNumberLookup --
 *
 *      Searches a join table for a value.  Calls SdbLuSearch to
 *      find the record, pulls out the required field, stores the
 *      record pointer in the cache.
 *
 * Results:
 *      Returns the numeric value.
 *
 * Side effects:
 *      Updates the cache.
 *
 *------------------------------------------------------------
 */
double
SdbNumberLookup(dbPtr, luHandlePtr, varIndex, rec)
     SdbFile                *dbPtr;
     SdbLookupHandle        *luHandlePtr;
     int                     varIndex;
     int                     rec;
{

    SdbLookup              *luPtr = luHandlePtr->luPtr;
    SdbTerminal           **cache;

    if (dbPtr->stamp != luHandlePtr->stamp) {
        cache = (SdbTerminal **) SdbLuSearch(dbPtr, luPtr->head,
                luHandlePtr->index,
                luPtr->numIndex, 0, rec);
        if (cache)
            luHandlePtr->cache = *cache;
        else
            luHandlePtr->cache = NULL;
        luHandlePtr->stamp = dbPtr->stamp;
    }
    if (luHandlePtr->cache)
        return ((luHandlePtr->cache)->value[varIndex]);
    else
        return (0.0);
}


/*
 *------------------------------------------------------------
 *
 * SdbPutNumberLookup --
 *
 *      Places a value in a join table.  Calls SdbLuSearch to
 *      find the record, inserts the required field, stores the
 *      record pointer in the cache.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Updates the cache.
 *
 *------------------------------------------------------------
 */
void
SdbPutNumberLookup(dbPtr, luHandlePtr, varIndex, rec, value)
     SdbFile                *dbPtr;
     SdbLookupHandle        *luHandlePtr;
     int                     varIndex;
     int                     rec;
     double                  value;
{

    SdbLookup              *luPtr = luHandlePtr->luPtr;
    SdbTerminal           **cache;

    if (dbPtr->stamp != luHandlePtr->stamp) {
        cache = (SdbTerminal **) SdbLuSearch(dbPtr, luPtr->head,
                luHandlePtr->index,
                luPtr->numIndex, 0, rec);
        if (cache)
            luHandlePtr->cache = *cache;
        else
            luHandlePtr->cache = NULL;
        luHandlePtr->stamp = dbPtr->stamp;
    }
    if (luHandlePtr->cache)
        (luHandlePtr->cache)->value[varIndex] = value;
}


/*
 *------------------------------------------------------------
 *
 * process_point --
 *
 *      Inserts a vertice into a yield table.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      May add more branches to the tree containing the table..
 *
 *------------------------------------------------------------
 */
static void
process_point(yield_set, path, values)
     SdbYieldSet            *yield_set;
     char                  **path;
     double                  values[];
{
    int                     k;
    SdbTree                *head = yield_set->tree,
                           *ptr;
    SdbCurve              **prev,
                           *curve;

#define MAX_PATH 10
    static char            *save_path[MAX_PATH];
    static SdbYieldSet     *save_set=NULL;
    static SdbCurve       **save_prev=NULL;

    /*
     * First check to see if there is a cached value for this
     * yield curve.  If so, this will speed things up a bit because
     * there will be no need to scan the tree or loop to the end of
     * the curve set.  The cache is valid if the yield set pointer
     * address is identical and if all the elements of the path
     * are also identical.  If so, then jump directly to the end
     * of the routine.
     */

    if (yield_set == save_set &&
	yield_set->numCats <= MAX_PATH) {
      for (k = 0; k < yield_set->numCats; k++) {
	if (strcmp(path[k], save_path[k]))
	  goto no_match;
      }
      prev = save_prev;
      goto got_match;
    }

    /*
     * No cached value. Search the tree for a matching value.
     * Create all of the branches as required along the way.
     */
  no_match:
    for (k = 0; k < yield_set->numCats; k++) {
        if (!strcmp(path[k], "*")) {
            if (head->wild)
                ptr = head->wild;
            else {
                ptr = head->wild = make_node(path[k]);
            }
            head = ptr;
        } else {
            if (!(ptr = (SdbTree *) tree_srch((tree **) & head->tree,
                                    comp_label1, path[k]))) {
                ptr = make_node(path[k]);
                tree_add((tree **) & head->tree, comp_label2, ptr,
                        SdbDelJoinTree);
            }
            head = ptr;
        }
    }

    /* 
     * At this stage we should be at the end of the tree. 
     * Remember the location, and then add then scan to the
     * end of the current vertice list.
     */
    save_set = yield_set;
    for (k = 0; k < yield_set->numCats; k++) 
      save_path[k] = path[k];

    curve = (SdbCurve *) ptr->terminal;
    prev = (SdbCurve **) &ptr->terminal;
    while (curve) {
        prev = &curve->next;
        curve = curve->next;
    }

    /* 
     * Ready to add the current vertice onto the end of the
     * yield list.  After adding it in, keep track of the end
     * of this list.
     */
  got_match:
    *prev = curve = (SdbCurve *) ckalloc(sizeof(SdbCurve));
    curve->values = values;
    curve->next = NULL;
    save_prev = &curve->next;

    return;
}



/*
 *------------------------------------------------------------
 *
 * SdbInterpolate --
 *
 *      Interpolate a value in a series of points
 *
 * Results:
 *      Returns the interpolate value.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static double
SdbInterpolate(point, index1, index2, x)
     SdbCurve               *point;
     int                     index1;
     int                     index2;
     double                  x;
{
    double                  a_x,
                            a_y,
                            b_x,
                            b_y,
                            y;

    y = 0.0;

    if (point) {

        a_x = point->values[index1];
        a_y = point->values[index2];

        b_x = a_x;
        b_y = a_y;

        while (((point = point->next) != NULL) && ((a_x < x) && (b_x < x))) {

            a_x = b_x;
            a_y = b_y;

            b_x = point->values[index1];
            b_y = point->values[index2];
        }

        if (x <= a_x)
            y = a_y;
        else if (x >= b_x)
            y = b_y;
        else if (a_x == b_x)
            y = a_y;
        else
            y = a_y + ((b_y - a_y) / (b_x - a_x)) * (x - a_x);
    }
    return (y);
}


/*
 *------------------------------------------------------------
 *
 * SdbYieldLookup --
 *
 *      Find a yield curve, then interpolate a value in the curve.
 *
 * Results:
 *      Returns the interpolated value.
 *
 * Side effects:
 *      Updates the cache.
 *
 *------------------------------------------------------------
 */
double
SdbYieldLookup(dbPtr, yield_data, rec, expr)
     SdbFile                *dbPtr;
     SdbYieldHandle         *yield_data;
     int                     rec;
     SdbNode                *expr;
{

    SdbYieldSet            *yield_set = yield_data->yieldSet;
    SdbCurve              **cache, **curve;
    SdbVariable           **list;
    unsigned int           *stamp;

    if (yield_data->list) {
      list = yield_data->list;
      stamp = &yield_data->stamp;
      cache = &yield_data->cache;
    }
    else {
      list = yield_set->list;
      stamp = &yield_set->stamp;
      cache = &yield_set->cache;
    }

    if (dbPtr->stamp != *stamp) {
        curve = (SdbCurve **) SdbLuSearch(dbPtr, yield_set->tree, list,
                yield_set->numCats, 0, rec);
	if (!curve)
	  *cache = NULL;
	else 
	  *cache = *curve;

        *stamp = dbPtr->stamp;
    }
    if (!*cache)
        return (0.0);

    return (SdbInterpolate(*cache, yield_data->xIndex,
                    yield_data->yIndex,
                    SdbEvaluate(dbPtr, rec, expr)));
}


/*
 *------------------------------------------------------------
 *
 * SdbGetCurve --
 *
 *      Retrieves the values that define a yield curve.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
SdbGetCurve(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    Tcl_HashEntry          *entryPtr;
    SdbYieldSet            *yield_set;
    char                  **listArgv;
    int                     listArgc;
    SdbCurve               *curve,
                           *next;
    int                     i;

    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    if (!(entryPtr = Tcl_FindHashEntry(dbPtr->yieldList, argv[0]))) {
        Tcl_AppendResult(interp, "curve set \"", argv[0], "\" not found", NULL);
        return TCL_ERROR;
    }
    yield_set = (SdbYieldSet *) Tcl_GetHashValue(entryPtr);
    if (Tcl_SplitList(interp, argv[1], &listArgc, &listArgv) != TCL_OK)
        return TCL_ERROR;
    if (listArgc != yield_set->numCats) {
        Tcl_AppendResult(interp, "wrong number of class variables", NULL);
        ckfree(listArgv);
        return TCL_ERROR;
    }
    curve = *SdbLuDirectSearch(dbPtr, yield_set->tree, listArgv,
            yield_set->numCats, 0);
    if (!curve) {
        Tcl_AppendResult(interp, "curve not found", NULL);
        ckfree(listArgv);
        return TCL_ERROR;
    }
    for (i = 0; i < yield_set->numValues; i++) {
        char                    temp[32];

	if (interp->result && *interp->result)
	  Tcl_AppendResult(interp, " ", NULL);
        Tcl_AppendResult(interp, "{", NULL);
        next = curve;
        while (next) {
            sprintf(temp, "%-.15g", next->values[i]);
            Tcl_AppendElement(interp, temp);
            next = next->next;
        }
        Tcl_AppendResult(interp, "}", NULL);
    }
    ckfree(listArgv);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbDefineCurve --
 *
 *      Defines the values that make up a yield curve.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      May create new branches in the tree.
 *
 *------------------------------------------------------------
 */
static int
SdbDefineCurve(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{

    Tcl_HashEntry          *entryPtr;
    SdbYieldSet            *yield_set;
    char                  **listArgv,
                          **valueArgv,
                         ***pointArgv;
    int                     listArgc,
                            valueArgc,
                           *pointArgc;
    SdbCurve              **ptr,
                           *curve,
                           *next;
    int                     i,
                            j;
    double                 *values;

    if (argc != 3) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    if (!(entryPtr = Tcl_FindHashEntry(dbPtr->yieldList, argv[0]))) {
        Tcl_AppendResult(interp, "curve set \"", argv[0], 
			 "\" not found", NULL);
        return TCL_ERROR;
    }
    yield_set = (SdbYieldSet *) Tcl_GetHashValue(entryPtr);
    if (Tcl_SplitList(interp, argv[1], &listArgc, &listArgv) != TCL_OK)
        return TCL_ERROR;
    if (listArgc != yield_set->numCats) {
        Tcl_AppendResult(interp, "wrong number of class variables", NULL);
        ckfree(listArgv);
        return TCL_ERROR;
    }
    if (Tcl_SplitList(interp, argv[2], &valueArgc, &valueArgv) != TCL_OK)
        return TCL_ERROR;
    if (valueArgc != yield_set->numValues) {
        Tcl_AppendResult(interp, "wrong number of curve lists", NULL);
        ckfree(listArgv);
        ckfree(valueArgv);
        return TCL_ERROR;
    }
    /* build an array of list pointers and split the lists */
    pointArgv = (char ***) ckalloc(yield_set->numValues * sizeof(char **));
    pointArgc = (int *) ckalloc(yield_set->numValues * sizeof(int));

    for (i = 0; i < yield_set->numValues; i++) {
        if ((Tcl_SplitList(interp, valueArgv[i], &pointArgc[i], &pointArgv[i])
                        != TCL_OK) ||
                (pointArgc[i] != pointArgc[0])) {
            ckfree(listArgv);
            ckfree(valueArgv);
            for (j = i - 1; j >= 0; j--)
                ckfree(pointArgv[j]);
            ckfree(pointArgv);
            if (pointArgc[i] != pointArgc[0])
                Tcl_AppendResult(interp, 
				 "number of points in curves not equal", NULL);
            ckfree(pointArgc);
            return TCL_ERROR;
        }
    }
    /* if this curve exists then get rid of it. */
    ptr = (SdbCurve **) SdbLuDirectSearch(dbPtr, yield_set->tree, listArgv,
            yield_set->numCats, 0);
    if (ptr) {
        curve = *ptr;
        *ptr = NULL;
        while (curve) {
            next = curve->next;
            ckfree(curve->values);
            ckfree(curve);
            curve = next;
        }
        *ptr = NULL;
    }
    /* Stash the index names into the hash table for safe keeping */
    for (i=0;i<listArgc;i++)
      listArgv[i]=GetIndices(dbPtr->htbl,listArgv[i],&j);

    /* loop through the points in the lists and add them into the
     * curve structures */
    for (i = 0; i < pointArgc[0]; i++) {
        values = (double *) ckalloc(yield_set->numValues * sizeof(double));

        for (j = 0; j < valueArgc; j++) {
            if (sscanf((pointArgv[j])[i], "%lf", &values[j]) != 1)
                values[j] = 0.0;
        }
        process_point(yield_set, listArgv, values);
    }

    /* clean up the temporary structures. */
    ckfree(listArgv);
    ckfree(valueArgv);
    for (i = 0; i < yield_set->numValues; i++)
        ckfree(pointArgv[i]);
    ckfree(pointArgv);
    ckfree(pointArgc);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * list_classes --
 *
 *      Helper function to scan a yield tree and find all of
 *      the classes that have been defined.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
list_classes(node)
     SdbTree                *node;
{
    int                     i;

    classList[currentDepth++] = node->label;
    if (node->terminal) {
        Tcl_AppendResult(currentInterp, "{", NULL);
        for (i = 0; i < currentDepth; i++)
            Tcl_AppendElement(currentInterp, classList[i]);
        Tcl_AppendResult(currentInterp, "} ", NULL);
    } else {
        if (node->tree)
            tree_trav((tree **) & node->tree, list_classes);
        if (node->wild)
            list_classes(node->wild);
    }
    currentDepth--;
    return (1);
}

/*
 *------------------------------------------------------------
 *
 * SdbCreateSets --
 *
 *      Creates a curve set definition
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *
 *------------------------------------------------------------
 */
static int
SdbCreateSets(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{
    char                   *set_name,
                          **indexV,
			  **valueV;
    int                     i,
                            indexC,
                            valueC;

    int                     new;
    SdbYieldSet            *yield_set;
    Tcl_HashEntry          *entryPtr;

    if (argc != 3) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    set_name = argv[0];

    if ((entryPtr = Tcl_FindHashEntry(dbPtr->yieldList, set_name))) {
        Tcl_AppendResult(interp, "curve set already exists", NULL);
        return TCL_ERROR;
    }

    /* 
     * Add an entry into the hash table, create the set structure,
     * and fill in the column headings
     */

    if (Tcl_SplitList(interp,argv[1],&indexC, &indexV) == TCL_ERROR) {
        Tcl_AppendResult(interp, "improper index field names", NULL);
        return TCL_ERROR;
    }
    if (indexC < 1) {
        Tcl_AppendResult(interp, "too few index names", NULL);
	ckfree(indexV);
        return TCL_ERROR;
    }
    if (Tcl_SplitList(interp,argv[2],&valueC, &valueV) == TCL_ERROR) {
        Tcl_AppendResult(interp, "improper value field names", NULL);
	ckfree(indexV);
        return TCL_ERROR;
    }
    if (valueC < 2) {
        Tcl_AppendResult(interp, "too few value names", NULL);
	ckfree(indexV);
	ckfree(valueV);
        return TCL_ERROR;
    }
      
    for (i = 0; i < indexC; i++) {
      if (!SdbFound(dbPtr->varPtr, indexV[i])) {
	Tcl_AppendResult(interp, "index field does not exist \"",
			 indexV[i], "\"", NULL);
	ckfree(indexV);
	ckfree(valueV);
	return TCL_ERROR;
      }
    }

    entryPtr = Tcl_CreateHashEntry(dbPtr->yieldList, set_name, &new);
    yield_set = (SdbYieldSet *) ckalloc(sizeof(SdbYieldSet));

    Tcl_SetHashValue(entryPtr, yield_set);

    yield_set->numCats = indexC;
    yield_set->numValues = valueC;
    yield_set->cats = (char **) ckalloc(yield_set->numCats * 
					sizeof(char *));
    yield_set->values = (char **) ckalloc(yield_set->numValues * 
					  sizeof(char *));
    
    for (i = 0; i < indexC; i++)
      yield_set->cats[i] = strdup(indexV[i]);
    for (i = 0; i < valueC; i++)
      yield_set->values[i] = strdup(valueV[i]);
    yield_set->tree = (SdbTree *) ckalloc(sizeof(SdbTree));
    yield_set->tree->label = NULL;
    yield_set->tree->tree = NULL;
    yield_set->tree->wild = NULL;
    yield_set->tree->terminal = NULL;
    
    yield_set->list =
      (SdbVariable **) ckalloc(yield_set->numCats * 
			       sizeof(SdbVariable *));
    ckfree(indexV);
    ckfree(valueV);
    
    return TCL_OK;

  }

/*
 *------------------------------------------------------------
 *
 * SdbLoadYields --
 *
 *      Loads up a yield curve set from definitions that
 *      are stored in a file.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      May create the yield curve set if it does not already
 *      exist.  May add branches onto the yield tree.
 *
 *------------------------------------------------------------
 */
static int
SdbLoadYields(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{
    char                   *filename,
                           *set_name,
                          **list;
    int                     i;

    SdbTextFile            *file;
    SdbYieldSet            *yield_set;
    Tcl_HashEntry          *entryPtr;
    double                 *values;

    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong number of arguments", NULL);
        return TCL_ERROR;
    }
    filename = argv[1];
    set_name = argv[0];

    if (!(file = SdbTextOpen(interp, filename, 8192))) {
        Tcl_AppendResult(interp, "file open failure: ", filename,
                (char *) NULL);
        return TCL_ERROR;
    }
    if (!(entryPtr = Tcl_FindHashEntry(dbPtr->yieldList, set_name))) {
        /* This is a new yield set.
         * Add an entry into the hash table, create the set structure,
         * and fill in the column headings
         */

        int                     new;

        if (file->split == 0) {
            Tcl_AppendResult(interp, "no class labels in yield file",
                    (char *) NULL);
            SdbTextClose(file);
            return TCL_ERROR;
        }
        if ((file->n - file->split) < 2) {
            Tcl_AppendResult(interp, "too few values in yield file",
                    (char *) NULL);
            SdbTextClose(file);
            return TCL_ERROR;
        }
        for (i = 0; i < file->split; i++) {
            if (!SdbFound(dbPtr->varPtr, file->field[i])) {
                Tcl_AppendResult(interp, "yield curve field does not exist \"",
                        file->field[i], "\"", NULL);
                SdbTextClose(file);
                return TCL_ERROR;
            }
        }

        entryPtr = Tcl_CreateHashEntry(dbPtr->yieldList, set_name, &new);
        yield_set = (SdbYieldSet *) ckalloc(sizeof(SdbYieldSet));

        Tcl_SetHashValue(entryPtr, yield_set);

        yield_set->numCats = file->split;
        yield_set->numValues = file->n - file->split;
        yield_set->cats = (char **) ckalloc(yield_set->numCats * 
					    sizeof(char *));
        yield_set->values = (char **) ckalloc(yield_set->numValues * 
					      sizeof(char *));

        for (i = 0; i < file->split; i++)
            yield_set->cats[i] = strdup(file->field[i]);
        for (i = file->split; i < file->n; i++)
            yield_set->values[i - file->split] = strdup(file->field[i]);
        yield_set->tree = (SdbTree *) ckalloc(sizeof(SdbTree));
	yield_set->tree->label = NULL;
	yield_set->tree->tree = NULL;
	yield_set->tree->wild = NULL;
	yield_set->tree->terminal = NULL;

        yield_set->list =
                (SdbVariable **) ckalloc(yield_set->numCats * 
					 sizeof(SdbVariable *));

    } else {
        /* This is an existing yield set file.  Check for
         * identical column headings.
         */

        yield_set = (SdbYieldSet *) Tcl_GetHashValue(entryPtr);

        if ((yield_set->numCats + yield_set->numValues) != file->n) {
            Tcl_AppendResult(interp, "wrong number of columns in yield file",
                    (char *) NULL);
            SdbTextClose(file);
            return TCL_ERROR;
        }
        for (i = 0; i < yield_set->numCats; i++) {
            if (!SdbTextReposition(file, yield_set->cats[i], i)) {
                Tcl_AppendResult(interp, "missing class label \"",
                        yield_set->cats[i],
                        "\" in yield file",
                        (char *) NULL);
                SdbTextClose(file);
                return TCL_ERROR;
            }
        }
        for (i = 0; i < yield_set->numValues; i++) {
            if (!SdbTextReposition(file, yield_set->values[i],
                            yield_set->numCats + i)) {
                Tcl_AppendResult(interp, "missing values \"",
                        yield_set->values[i],
                        "\" in yield file",
                        (char *) NULL);
                SdbTextClose(file);
                return TCL_ERROR;
            }
        }

    }

    /* update the list of where each variable is located */
    for (i = 0; i < yield_set->numCats; i++) {
        yield_set->list[i] = SdbFound(dbPtr->varPtr, yield_set->cats[i]);
    }

    list = (char **) ckalloc(yield_set->numCats * sizeof(char *));

    while (SdbTextGet(file)) {

        for (i = 0; i < yield_set->numCats; i++) {
            int                     new;

            entryPtr = Tcl_CreateHashEntry(dbPtr->htbl, file->list[i], &new);
            list[i] = Tcl_GetHashKey(dbPtr->htbl, entryPtr);
        }
        values = (double *) ckalloc(yield_set->numValues * sizeof(double));

        for (i = 0; i < (yield_set->numValues); i++) {
            if (!sscanf(file->list[yield_set->numCats + i], "%lf",
                            &values[i])) {
                Tcl_AppendResult(interp,
                        "invalid value in yield file \"",
                        file->list[yield_set->numCats + i], "\"\n",
                        NULL);
                values[i] = 0.0;
            }
        }

        process_point(yield_set, list, values);

    }

    yield_set->stamp = 0;
    yield_set->cache = NULL;
    SdbTextClose(file);
    ckfree(list);
    SdbReOptimize(dbPtr);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbDelYieldTree --
 *
 *      Deletes all the branches in a yield curve tree.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Frees memory.
 *
 *------------------------------------------------------------
 */
static void
SdbDelYieldTree(node)
     SdbTree                *node;
{
    if (!node)
        return;
    if (node->tree)
        tree_mung((tree **) & node->tree, SdbDelYieldTree);
    if (node->wild)
        SdbDelYieldTree(node->wild);
    if (node->terminal) {
        SdbCurve               *curve = (SdbCurve *) node->terminal;
        SdbCurve               *next;

        while (curve) {
            next = curve->next;
            ckfree(curve->values);
            ckfree(curve);
            curve = next;
        }
    }
    ckfree(node);
    return;
}


/*
 *------------------------------------------------------------
 *
 * SdbDeleteYieldSet --
 *
 *      Deletes a yield set and all the storage that it has 
 *      allocated.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Frees memory.
 *
 *------------------------------------------------------------
 */
void
SdbDeleteYieldSet(yield_set)
     SdbYieldSet            *yield_set;
{
    int                     i;

    if (!yield_set)
        return;

    for (i = 0; i < yield_set->numCats; i++)
        ckfree(yield_set->cats[i]);
    ckfree(yield_set->cats);
    for (i = 0; i < yield_set->numValues; i++)
        ckfree(yield_set->values[i]);
    ckfree(yield_set->values);
    ckfree(yield_set->list);
    SdbDelYieldTree(yield_set->tree);
    ckfree(yield_set);
}


/*
 *------------------------------------------------------------
 *
 * SdbJoinDeleteRec --
 *
 *      Deletes a record from a join tree.
 *
 * Results:
 *      Returns a stadard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
SdbJoinDeleteRec(interp, dbPtr, luPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     SdbLookup              *luPtr;
     int                     argc;
     char                  **argv;
{
    char                  **indexV;
    int                     indexC;

    if (argc != 1) {
        Tcl_AppendResult(interp, "wrong number of arguments: insert {index}", 
			 NULL);
        return TCL_ERROR;
    }
    /* check the arguments to see if they are proper */
    if (Tcl_SplitList(interp, argv[0], &indexC, &indexV) != TCL_OK) {
        return TCL_ERROR;
    }
    if (indexC != luPtr->numIndex) {
        Tcl_AppendResult(interp, "wrong number of index elements", NULL);
        ckfree(indexV);
        return TCL_ERROR;
    }
    rec_remove(luPtr->head, indexV, indexC);
    ckfree(indexV);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbJoinInsert --
 *
 *      Inserts a record into a join tree.
 *
 * Results:
 *      Returns a stadard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
SdbJoinInsert(interp, dbPtr, luHandlePtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     SdbLookupHandle        *luHandlePtr;
     int                     argc;
     char                  **argv;
{
    int                     i,
                            new;
    char                  **indexV,
                          **codeV,
                          **valueV;
    int                     indexC,
                            codeC,
                            valueC;
    SdbTerminal             newNode;
    char                  **path,
                          **codes;
    double                 *values;
    Tcl_HashTable          *htbl;
    SdbLookup              *luPtr = luHandlePtr->luPtr;

    if (argc != 3) {
        Tcl_AppendResult(interp, "wrong number of arguments: insert {index} {codes} {values}", NULL);
        return TCL_ERROR;
    }
    /* check the arguments to see if they are proper */
    if (Tcl_SplitList(interp, argv[0], &indexC, &indexV) != TCL_OK) {
        return TCL_ERROR;
    }
    if (Tcl_SplitList(interp, argv[1], &codeC, &codeV) != TCL_OK) {
        ckfree(indexV);
        return TCL_ERROR;
    }
    if (Tcl_SplitList(interp, argv[2], &valueC, &valueV) != TCL_OK) {
        ckfree(indexV);
        ckfree(codeV);
        return TCL_ERROR;
    }
    if (indexC != luPtr->numIndex ||
            codeC != luPtr->numCodes ||
            valueC != luPtr->numVals) {
        if (indexC != luPtr->numIndex)
            Tcl_AppendResult(interp, "wrong number of index elements", NULL);
        if (codeC != luPtr->numCodes)
            Tcl_AppendResult(interp, "wrong number of code elements", NULL);
        if (valueC != luPtr->numVals)
            Tcl_AppendResult(interp, "wrong number of value elements", NULL);
        ckfree(indexV);
        ckfree(codeV);
        ckfree(valueV);
        return TCL_ERROR;
    }
    path = (char **) ckalloc(indexC * sizeof(char *));
    codes = (char **) ckalloc(codeC * sizeof(char *));
    values = (double *) ckalloc(valueC * sizeof(double));

    /* check if this is a remotely connected database.  If so then
     * stash the strings in the remote dictionary.  When the
     * current table is closed, these strings will still exist. */
    if (luHandlePtr->remote)
        htbl = luHandlePtr->remote;
    else
        htbl = dbPtr->htbl;

    for (i = 0; i < indexC; i++)
        path[i] = (Tcl_CreateHashEntry(htbl, indexV[i], &new))->key.string;
    for (i = 0; i < codeC; i++)
        codes[i] = (Tcl_CreateHashEntry(htbl, codeV[i], &new))->key.string;
    for (i = 0; i < valueC; i++) {
        values[i] = 0;
        sscanf(valueV[i], "%lf", &values[i]);
    }
    ckfree(indexV);
    ckfree(codeV);
    ckfree(valueV);

    newNode.codes = codes;
    newNode.value = values;

    place(luPtr->head, path, indexC, &newNode);
    ckfree(codes);
    ckfree(path);
    ckfree(values);
    return TCL_OK;
}



/*
 *------------------------------------------------------------
 *
 * trav --
 *
 * A helper routine for the 'saveHelper' routine.  This is 
 * called recursively for all nodes at a given level within
 * the search tree.  It scans all nodes, calling back to
 * 'saveHelper' at each node.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      None
 *
 *------------------------------------------------------------
 */
static void
trav(interp, output, ppr_tree, luPtr, path, i)
     Tcl_Interp             *interp;
     FILE                   *output;
     tree                  **ppr_tree;
     SdbLookup              *luPtr;
     char                  **path;
     int                     i;
{
    if (!*ppr_tree)
        return;
    trav(interp, output, &(**ppr_tree).tree_l, luPtr, path, i);
    path[i] = ((SdbTree *) (**ppr_tree).tree_p)->label;
    saveHelper(interp, output, (**ppr_tree).tree_p, luPtr, path, i + 1);
    trav(interp, output, &(**ppr_tree).tree_r, luPtr, path, i);
    return;
}


/*
 *------------------------------------------------------------
 *
 * saveHelper --
 *
 * A helper routine for saving join files.  This routine is 
 * called recursively for each level of the index, and builds
 * up a list called 'path' that contains the search key.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      writes to a file.  Appends to the result string.
 *
 *------------------------------------------------------------
 */
static void
saveHelper(interp, output, head, luPtr, path, i)
     Tcl_Interp             *interp;
     FILE                   *output;
     SdbTree                *head;
     SdbLookup              *luPtr;
     char                  **path;
     int                     i;
{
    int                     j;

    if (!head)
        return;
    if (head->terminal) {
        Tcl_AppendElement(interp, "DATA");
        for (j = 0; j < luPtr->numIndex; j++)
            Tcl_AppendElement(interp, path[j]);
        for (j = 0; j < luPtr->numCodes; j++)
            Tcl_AppendElement(interp, (head->terminal)->codes[j]);
        Tcl_AppendElement(interp, ":");
        for (j = 0; j < luPtr->numVals; j++) {
            char                    temp[64];

            sprintf(temp, "%-.15g", head->terminal->value[j]);
            Tcl_AppendElement(interp, temp);
        }
        fprintf(output, "%s\n", interp->result);
        Tcl_ResetResult(interp);
        return;
    }
    if (head->tree)
        trav(interp, output, (tree**)&head->tree, luPtr, path, i);
    if (head->wild) {
        path[i] = "*";
        saveHelper(interp, output, head->wild, luPtr, path, i + 1);
    }
}


/*
 *------------------------------------------------------------
 *
 * SdbJoinSave --
 *
 *      Saves the contents of a join tree out ot a file.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      Writes a file.
 *
 *------------------------------------------------------------
 */
static int
SdbJoinSave(interp, dbPtr, luPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     SdbLookup              *luPtr;
     int                     argc;
     char                  **argv;
{
    int                     i;
    FILE                   *output;
    char                  **path;

    if (argc != 1) {
        Tcl_AppendResult(interp, "wrong number of arguments: save file", NULL);
        return TCL_ERROR;
    }
    output = fopen(argv[0], "w");
    if (!output) {
        Tcl_AppendResult(interp, "could not open output file \"",
                argv[0], "\"", NULL);
        return TCL_ERROR;
    }
    /* print out the variable header record */
    Tcl_ResetResult(interp);
    Tcl_AppendElement(interp, "VAR");
    for (i = 0; i < luPtr->numIndex; i++)
        Tcl_AppendElement(interp, luPtr->indexNames[i]);
    for (i = 0; i < luPtr->numCodes; i++)
        Tcl_AppendElement(interp, luPtr->codeNames[i]);
    Tcl_AppendElement(interp, ":");
    for (i = 0; i < luPtr->numVals; i++)
        Tcl_AppendElement(interp, luPtr->valueNames[i]);
    fprintf(output, "%s\n", interp->result);
    Tcl_ResetResult(interp);

    path = (char **) ckalloc(luPtr->numIndex * sizeof(char **));

    saveHelper(interp, output, luPtr->head, luPtr, path, 0);
    ckfree(path);
    fclose(output);
    Tcl_AppendResult(interp, argv[0], NULL);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbJoinInfo --
 *
 *      Obtains information about a join table
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
static int
SdbJoinInfo(interp, dbPtr, luPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     SdbLookup              *luPtr;
     int                     argc;
     char                  **argv;
{
    int                     i;

    if (argc != 1) {
        Tcl_AppendResult(interp, "wrong number of arguments: info option",
			 NULL);
        return TCL_ERROR;
    }
    if (!strcmp("index", argv[0])) {
        for (i = 0; i < luPtr->numIndex; i++)
            Tcl_AppendElement(interp, luPtr->indexNames[i]);
        return TCL_OK;
    } else if (!strcmp("codes", argv[0])) {
        for (i = 0; i < luPtr->numCodes; i++)
            Tcl_AppendElement(interp, luPtr->codeNames[i]);
        return TCL_OK;
    } else if (!strcmp("values", argv[0])) {
        for (i = 0; i < luPtr->numVals; i++)
            Tcl_AppendElement(interp, luPtr->valueNames[i]);
        return TCL_OK;
    } else {
        Tcl_AppendResult(interp, "bad option \"", argv[0],
                "\": should be codes, index or values", NULL);
        return TCL_ERROR;
    }
}


/*
 *------------------------------------------------------------
 *
 * SdbJoinConnect --
 *
 *      Makes a connection to a join table in a remote database
 *      (in a different sdb database).  The table will look like
 *      a normal join table, except only one copy will exist and
 *      values will be shared between the two tables.
 *
 * Results:
 *      Returns a standard TCL error code.
 *
 * Side effects:
 *      Increments the reference counts in the remote table,
 *      adds new variables to this database, creates a new
 *      SdbLuHandle structure.
 *
 *------------------------------------------------------------
 */
static int
SdbJoinConnect(interp, dbPtr, argc, argv)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
{
    Tcl_HashEntry          *entryPtr;
    SdbLookupHandle        *newLuHandlePtr,
                           *srcLuHandlePtr;
    SdbLookup              *srcLuPtr;
    int                     i,
                            new;
    char                   *srcTable,
                           *srcDb;
    SdbFile                *srcDbPtr;
    SdbVariable            *vars;

    /* start parsing the arguments to make sure that things are correct */
    if (argc != 2) {
        Tcl_AppendResult(interp, 
			 "wrong number of arguments: connect luName source", NULL);
        return TCL_ERROR;
    }
    /* destination table name already in use */
    entryPtr = Tcl_FindHashEntry(dbPtr->joinList, argv[0]);
    if (entryPtr) {
        Tcl_AppendResult(interp, "join table \"", argv[0], 
			 "\" already exists", NULL);
        return TCL_ERROR;
    }
    /* bad source specifier */
    if (!strchr(argv[1], '@')) {
        Tcl_AppendResult(interp,
                "source table name not in correct format (lu@db) \"",
                argv[1], "\"", NULL);
        return TCL_ERROR;
    }
    srcTable = strtok(argv[1], "@");
    srcDb = strtok(NULL, "");

    /* missing source table name */
    if (!srcTable) {
        Tcl_AppendResult(interp,
                "missing source table name in specifier \"",
                argv[1], "\"", NULL);
        return TCL_ERROR;
    }
    /* missing source database name */
    if (!srcDb) {
        Tcl_AppendResult(interp,
                "missing source database name in specifier \"",
                argv[1], "\"", NULL);
        return TCL_ERROR;
    }
    /* trap another stupid error */
    if (strcmp(srcDb, dbPtr->dbname) == 0) {
        Tcl_AppendResult(interp,
                "not allowed to connect to self \"",
                srcDb, "\"", NULL);
        return TCL_ERROR;
    }
    /* locate the pointer to the source database */
    entryPtr = Tcl_FindHashEntry(&sdb_HashTable, srcDb);
    if (!entryPtr) {
        Tcl_AppendResult(interp,
                "source database not found \"",
                srcDb, "\"", NULL);
        return TCL_ERROR;
    }
    srcDbPtr = (SdbFile *) Tcl_GetHashValue(entryPtr);

    /* locate the table within the source database */
    entryPtr = Tcl_FindHashEntry(srcDbPtr->joinList, srcTable);
    if (!entryPtr) {
        Tcl_AppendResult(interp,
                "source table not found in remote database\"",
                srcTable, "@", srcDb, "\"", NULL);
        return TCL_ERROR;
    }
    srcLuHandlePtr = (SdbLookupHandle *) Tcl_GetHashValue(entryPtr);
    srcLuPtr = srcLuHandlePtr->luPtr;

    /* match the contents of the remote table to the current
     * table.  Check that indexing columns are available and
     * that the added columns do not already exist */

    for (i = 0; i < srcLuPtr->numIndex; i++) {
        if (!(vars = SdbFound(dbPtr->varPtr, srcLuPtr->indexNames[i]))) {
            Tcl_AppendResult(interp,
                    "index variable \"", srcLuPtr->indexNames[i],
                    "\" not found in current database", NULL);
            return TCL_ERROR;
        }
    }
    for (i = 0; i < srcLuPtr->numCodes; i++) {
        if ((vars = SdbFound(dbPtr->varPtr, srcLuPtr->codeNames[i]))) {
            Tcl_AppendResult(interp,
                    "added variable \"", srcLuPtr->indexNames[i],
                    "\" already exists in current database", NULL);
            return TCL_ERROR;
        }
    }
    for (i = 0; i < srcLuPtr->numVals; i++) {
        if ((vars = SdbFound(dbPtr->varPtr, srcLuPtr->valueNames[i]))) {
            Tcl_AppendResult(interp,
                    "added variable \"", srcLuPtr->indexNames[i],
                    "\" already exists in current database", NULL);
            return TCL_ERROR;
        }
    }

    /* So far so good.  It looks like it is safe to connect to
     * the remote table.  Create a new luPtr structure, 
     * create new variable entries for this database, and
     * then rebuild the database structure. */

    /* build the new lu pointer */
    newLuHandlePtr = (SdbLookupHandle *) ckalloc(sizeof(SdbLookupHandle));

    newLuHandlePtr->luPtr = srcLuPtr;
    newLuHandlePtr->cache = NULL;
    newLuHandlePtr->index =
            (SdbVariable **) ckalloc(srcLuPtr->numIndex * 
				     sizeof(SdbVariable *));

    newLuHandlePtr->remote = srcDbPtr->htbl;
    newLuHandlePtr->referenceCount = 0;
    newLuHandlePtr->stamp = 0;
    newLuHandlePtr->name = GetIndices(dbPtr->htbl, argv[0], &new);

    /* record locations of the index variables */
    for (i = 0; i < srcLuPtr->numIndex; i++)
        newLuHandlePtr->index[i] = SdbFound(dbPtr->varPtr, 
					    srcLuPtr->indexNames[i]);

    /* insert the added variables into the current */
    for (i = 0; i < srcLuPtr->numCodes; i++) {
        if (!SdbInsertLookup(dbPtr, srcLuPtr->codeNames[i],
                        SDB_LOOKUP | SDB_CODE | SDB_OUTPUT, 
			     newLuHandlePtr, i)) {
            Tcl_AppendResult(interp, "failed to insert variable \"",
                    srcLuPtr->codeNames[i], "\" into table", NULL);
        } else {
            srcLuPtr->referenceCount++;
            newLuHandlePtr->referenceCount++;
        }
    }
    for (i = 0; i < srcLuPtr->numVals; i++) {
        if (!SdbInsertLookup(dbPtr, srcLuPtr->valueNames[i],
                        SDB_LOOKUP | SDB_NUMBER | SDB_OUTPUT, 
			     newLuHandlePtr, i)) {
            Tcl_AppendResult(interp, "failed to insert variable \"",
                    srcLuPtr->valueNames[i], "\" into table", NULL);
        } else {
            srcLuPtr->referenceCount++;
            newLuHandlePtr->referenceCount++;
        }
    }

    /* add an entry into the join table catalog */
    entryPtr = Tcl_CreateHashEntry(dbPtr->joinList, argv[0], &new);
    Tcl_SetHashValue(entryPtr, newLuHandlePtr);

    /* rebuild the links */
    SdbReOptimize(dbPtr);

    Tcl_AppendResult(dbPtr->interp,argv[0],NULL);
    return TCL_OK;
}
